# noqa: D100
from __future__ import annotations

import cf_xarray  # noqa: F401, pylint: disable=unused-import
import numpy as np
import xarray

from xclim.core.units import convert_units_to, declare_units

# Frequencies : YS: year start, QS-DEC: seasons starting in december, MS: month start
# See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html

# -------------------------------------------------- #
# ATTENTION: ASSUME ALL INDICES WRONG UNTIL TESTED ! #
# -------------------------------------------------- #

__all__ = [
    "jetstream_metric_woollings",
]


@declare_units(ua="[speed]")
def jetstream_metric_woollings(
    ua: xarray.DataArray,
) -> tuple[xarray.DataArray, xarray.DataArray]:
    """Strength and latitude of jetstream.

    Identify latitude and strength of maximum smoothed zonal wind speed in the region from 15 to 75°N and -60 to 0°E,
    using the formula outlined in :cite:p:`woollings_variability_2010`. Wind is smoothened using a Lanczos filter
    approach.

    Warnings
    --------
    This metric expects eastward wind component (u) to be on a regular grid (i.e. Plate Carree, 1D lat and lon)

    Parameters
    ----------
    ua : xarray.DataArray
        Eastward wind component (u) at between 750 and 950 hPa.

    Returns
    -------
    (xarray.DataArray, xarray.DataArray)
        Daily time series of latitude of jetstream and daily time series of strength of jetstream.

    References
    ----------
    :cite:cts:`woollings_variability_2010`
    """
    # Select longitudes in the -60 to 0°E range
    lon = ua.cf["longitude"]
    ilon = (lon >= 300) * (lon <= 360) + (lon >= -60) * (lon <= 0)
    if not ilon.any():
        raise ValueError(
            "Make sure the grid includes longitude values in a range between -60 and 0°E."
        )

    # Select latitudes in the 15 to 75°N range
    lat = ua.cf["latitude"]
    ilat = (lat >= 15) * (lat <= 75)
    if not ilat.any():
        raise ValueError(
            "Make sure the grid includes latitude values in a range between 15 and 75°N."
        )

    # Select levels between 750 and 950 hPa
    pmin = convert_units_to("750 hPa", ua.cf["vertical"])
    pmax = convert_units_to("950 hPa", ua.cf["vertical"])

    p = ua.cf["vertical"]
    ip = (p >= pmin) * (p <= pmax)
    if not ip.any():
        raise ValueError(
            "Make sure the grid includes pressure values in a range between 750 and 950 hPa."
        )

    ua = ua.cf.sel(
        vertical=ip,
        latitude=ilat,
        longitude=ilon,
    )

    zonal_mean = ua.cf.mean(["vertical", "longitude"])

    # apply Lanczos filter - parameters are hard-coded following those used in Woollings (2010)
    filter_freq = 10
    window_size = 61
    cutoff = 1 / filter_freq
    if ua.time.size <= filter_freq or ua.time.size <= window_size:
        raise ValueError(
            f"Time series is too short to apply 61-day Lanczos filter (got a length of  {ua.time.size})"
        )

    # compute low-pass filter weights
    lanczos_weights = _compute_low_pass_filter_weights(
        window_size=window_size, cutoff=cutoff
    )
    # apply the filter
    ua_lf = (
        zonal_mean.rolling(time=window_size, center=True)
        .construct("window")
        .dot(lanczos_weights)
    )

    # Get latitude & eastward wind component units
    lat_name = ua.cf["latitude"].name
    lat_units = ua.cf["latitude"].units
    ua_units = ua.units

    jetlat = ua_lf.cf.idxmax(lat_name).rename("jetlat").assign_attrs(units=lat_units)
    jetstr = ua_lf.cf.max(lat_name).rename("jetstr").assign_attrs(units=ua_units)
    return jetlat, jetstr


def _compute_low_pass_filter_weights(
    window_size: int, cutoff: float
) -> xarray.DataArray:
    order = ((window_size - 1) // 2) + 1
    nwts = 2 * order + 1
    w = np.zeros([nwts])
    n = nwts // 2
    w[n] = 2 * cutoff
    k = np.arange(1.0, n)
    sigma = np.sin(np.pi * k / n) * n / (np.pi * k)
    firstfactor = np.sin(2.0 * np.pi * cutoff * k) / (np.pi * k)
    w[n - 1 : 0 : -1] = firstfactor * sigma
    w[n + 1 : -1] = firstfactor * sigma

    lanczos_weights = xarray.DataArray(w[0 + (window_size % 2) : -1], dims=["window"])
    return lanczos_weights



# Tests taken from flyingpigeon on Nov 2020
from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_almost_equal
from scipy import integrate, stats
from sklearn import datasets

import xclim.analog as xca


def matlab_sample(n=30):
    """
    In some of the following tests I'm using Matlab code written by Patrick
    Grenier for the paper "An Assessment of Six Dissimilarity Metrics for
    Climate Analogs" to compare against the functions here. The sample
    created here is identical to the sample used to drive the Matlab code.

    Parameters
    ----------
    n : int
      Sample size.

    Returns
    -------
    2D array, 2D array
       Synthetic samples (3, n)
    """
    z = 1.0 * (np.arange(n) + 1) / n - 0.5

    x = np.vstack([z * 2 + 30, z * 3 + 40, z]).T

    y = np.vstack([z * 2.2 + 31, z[::-1] * 2.8 + 38, z * 1.1]).T

    return x, y


@pytest.fixture
def exact_randn(random):
    def _randn(mean, std, shape):
        """Return a random normal sample with exact mean and standard deviation."""
        r = random.standard_normal(shape)
        r1 = r / r.std(0, ddof=1) * np.array(std)
        return r1 - r1.mean(0) + np.array(mean)

    return _randn


def test_exact_randn(exact_randn):
    mu, std = [2, 3], [1, 2]
    r = exact_randn(mu, std, [10, 2])
    assert_almost_equal(r.mean(0), mu)
    assert_almost_equal(r.std(0, ddof=1), std)


@pytest.mark.slow
@pytest.mark.parametrize("method", xca.metrics.keys())
def test_spatial_analogs(method, open_dataset):
    diss = open_dataset("SpatialAnalogs/dissimilarity.nc")
    data = open_dataset("SpatialAnalogs/indicators.nc")

    target = data.sel(lat=46.1875, lon=-72.1875, time=slice("1970", "1990"))
    candidates = data.sel(time=slice("1970", "1990"))

    out = xca.spatial_analogs(target, candidates, method=method)
    # Special case since scikit-learn updated to 1.2.0 (and again at 1.3)
    if method == "friedman_rafsky":
        diss[method][42, 105] = np.nan
        out[42, 105] = np.nan
    np.testing.assert_allclose(diss[method], out, rtol=1e-3, atol=1e-3)


def test_unsupported_spatial_analog_method(open_dataset):
    method = "KonMari"

    data = open_dataset("SpatialAnalogs/indicators.nc")
    target = data.sel(lat=46.1875, lon=-72.1875, time=slice("1970", "1990"))
    candidates = data.sel(time=slice("1970", "1990"))

    match_statement = f"Method `KonMari` is not implemented. Available methods are: {','.join(xca.metrics.keys())}"

    with pytest.raises(ValueError, match=match_statement):
        xca.spatial_analogs(target, candidates, method=method)


def test_spatial_analogs_multi_index(open_dataset):
    # Test multi-indexes
    diss = open_dataset("SpatialAnalogs/dissimilarity.nc")
    data = open_dataset("SpatialAnalogs/indicators.nc")

    target = data.sel(lat=46.1875, lon=-72.1875, time=slice("1970", "1990"))
    candidates = data.sel(time=slice("1970", "1990"))

    target_stacked = target.stack(sample=["time"])
    candidates_stacked = candidates.stack(sample=["time"])

    method = "seuclidean"
    out = xca.spatial_analogs(
        target_stacked, candidates_stacked, dist_dim="sample", method=method
    )
    np.testing.assert_allclose(diss[method], out, rtol=1e-3, atol=1e-3)

    # Check that it works as well when time dimensions don't have the same length.
    candidates = data.sel(time=slice("1970", "1991"))
    xca.spatial_analogs(
        target_stacked, candidates_stacked, dist_dim="sample", method=method
    )


class TestSEuclidean:
    def test_simple(self, exact_randn):
        d = 2
        n, m = 25, 30

        x = exact_randn(0, 1, (n, d))
        y = exact_randn([1, 2], 1, (m, d))
        dm = xca.seuclidean(x, y)
        assert_almost_equal(dm, np.hypot(1, 2), 2)

        # Variance of the candidate sample does not affect answer.
        x = exact_randn(0, 1, (n, d))
        y = exact_randn([1, 2], 2, (m, d))
        dm = xca.seuclidean(x, y)
        assert_almost_equal(dm, np.hypot(1, 2), 2)

    def test_compare_with_matlab(self):
        x, y = matlab_sample()
        dm = xca.seuclidean(x, y)
        assert_almost_equal(dm, 2.8463, 4)


class TestNN:
    def test_simple(self, random):
        d = 2
        n, m = 200, 200
        x = random.standard_normal((n, d))
        y = random.standard_normal((m, d))

        # Almost identical samples
        dm = xca.nearest_neighbor(x + 0.001, x)
        assert_almost_equal(dm, 0, 2)

        # Same distribution but mixed
        dm = xca.nearest_neighbor(x, y)
        assert_almost_equal(dm, 0.5, 1)

        # Two completely different distributions
        dm = xca.nearest_neighbor(x + 10, y)
        assert_almost_equal(dm, 1, 2)

    def test_compare_with_matlab(self):
        x, y = matlab_sample()
        dm = xca.nearest_neighbor(x, y)
        assert_almost_equal(dm, 1, 4)


class TestZAE:
    def test_simple(self, random):
        d = 2
        n = 200
        # m = 200
        x = random.standard_normal((n, d))
        # y = random.standard_normal(m, d)

        # Almost identical samples
        dm = xca.zech_aslan(x + 0.001, x)
        assert dm < 0

    def test_compare_with_matlab(self):
        x, y = matlab_sample()
        dm = xca.zech_aslan(x, y)
        assert_almost_equal(dm, 0.77802, 4)


class TestFR:
    def test_simple(self):
        # Over these 7 points, there are 2 with edges within the same sample.
        # [1,2]-[2,2] & [3,2]-[4,2]
        # |
        # |   x
        # | o o x x
        # | x  o
        # |_ _ _ _ _ _ _
        x = np.array([[1, 2], [2, 2], [3, 1]])
        y = np.array([[1, 1], [2, 4], [3, 2], [4, 2]])

        dm = xca.friedman_rafsky(x, y)
        assert_almost_equal(dm, 2.0 / 7, 3)

    def test_compare_with_matlab(self):
        x, y = matlab_sample()
        dm = xca.friedman_rafsky(x, y)
        assert_almost_equal(dm, 0.96667, 4)


class TestKS:
    def test_1D_ks_2samp(self, random):
        # Compare with scipy.stats.ks_2samp
        x = random.standard_normal(50) + 1
        y = random.standard_normal(50)
        s, _p = stats.ks_2samp(x, y)
        dm = xca.kolmogorov_smirnov(x, y)
        assert_almost_equal(dm, s, 3)

    def test_compare_with_matlab(self):
        x, y = matlab_sample()
        dm = xca.kolmogorov_smirnov(x, y)
        assert_almost_equal(dm, 0.96667, 4)


def analytical_KLDiv(p, q):
    """Return the Kullback-Leibler divergence between two distributions.

    Parameters
    ----------
    p, q : scipy.frozen_rv
      Scipy frozen distribution instances, e.g. stats.norm(0,1)

    Returns
    -------
    out : float
      The Kullback-Leibler divergence computed by numerically integrating
      p(x)*log(p(x)/q(x)).
    """

    def func(x):
        return p.pdf(x) * np.log(p.pdf(x) / q.pdf(x))

    a = 1e-5
    return integrate.quad(func, max(p.ppf(a), q.ppf(a)), min(p.isf(a), q.isf(a)))[0]


@pytest.mark.slow
class TestKLDIV:
    #
    def test_against_analytic(self, random):
        p = stats.norm(2, 1)
        q = stats.norm(2.6, 1.4)

        ra = analytical_KLDiv(p, q)

        N = 10000
        # x, y = p.rvs(N, random_state=random), q.rvs(N, random_state=random)

        re = xca.kldiv(p.rvs(N, random_state=random), q.rvs(N, random_state=random))

        assert_almost_equal(re, ra, 1)

    def test_accuracy(self, random):
        p = stats.norm(0, 1)
        q = stats.norm(0.2, 0.9)

        k = np.arange(1, 16)

        out = []
        n = 500
        for _i in range(500):
            out.append(
                xca.kldiv(
                    p.rvs(n, random_state=random), q.rvs(n, random_state=random), k=k
                )
            )
        out = np.array(out)

        # Compare with analytical value
        err = out - analytical_KLDiv(p, q)

        m = err.mean(0)
        assert_almost_equal(np.mean(m[0:2]), 0, 2)

    #
    def test_different_sample_size(self, random):
        p = stats.norm(2, 1)
        q = stats.norm(2.6, 1.4)

        ra = analytical_KLDiv(p, q)

        n = 6000
        # Same sample size for x and y
        re = [
            xca.kldiv(p.rvs(n, random_state=random), q.rvs(n, random_state=random))
            for i in range(30)
        ]
        assert_almost_equal(np.mean(re), ra, 2)

        # Different sample sizes
        re = [
            xca.kldiv(p.rvs(n * 2, random_state=random), q.rvs(n, random_state=random))
            for i in range(30)
        ]
        assert_almost_equal(np.mean(re), ra, 2)

        re = [
            xca.kldiv(p.rvs(n, random_state=random), q.rvs(n * 2, random_state=random))
            for i in range(30)
        ]
        assert_almost_equal(np.mean(re), ra, 2)

    #
    def test_mvnormal(self, random):
        """Compare the results to the figure 2 in the paper."""
        n = 30000
        p = random.normal(0, 1, size=(n, 2))
        q = random.multivariate_normal([0.5, -0.5], [[0.5, 0.1], [0.1, 0.3]], size=n)

        assert_almost_equal(xca.kldiv(p, q), 1.39, 1)
        assert_almost_equal(xca.kldiv(q, p), 0.62, 1)


def test_szekely_rizzo():
    iris = pd.DataFrame(datasets.load_iris().data)

    # first 80 against last 70
    x = iris.iloc[:80, :].to_xarray().to_array().T
    y = iris.iloc[80:, :].to_xarray().to_array().T

    np.testing.assert_allclose(
        xca.szekely_rizzo(x, y, standardize=False), 116.1987, atol=5e-5
    )

    # first 50 against last 100
    x = iris.iloc[:50, :].to_xarray().to_array().T
    y = iris.iloc[50:, :].to_xarray().to_array().T

    np.testing.assert_allclose(
        xca.szekely_rizzo(x, y, standardize=False), 199.6205, atol=5e-5
    )



from __future__ import annotations

import numpy as np
import xarray as xr


def _concat_hist(da: xr.DataArray, **hist) -> xr.DataArray:
    r"""Concatenate historical scenario with future scenarios along the time dimension.

    Parameters
    ----------
    da : xr.DataArray
        Input data where the historical scenario is stored alongside other, future, scenarios.
    \*\*hist : dict
        Mapping of the scenario dimension name to the historical scenario coordinate, e.g. `scenario="historical"`.

    Returns
    -------
    xr.DataArray
        Data with the historical scenario is stacked in time before each one of the other scenarios.

    Notes
    -----
    Data goes from:

        +------------+----------------------------------+
        | scenario   | time                             |
        +============+==================================+
        | historical | ``hhhhhhhhhhhhhhhh------------`` |
        +------------+----------------------------------+
        | ssp245     | ``----------------111111111111`` |
        +------------+----------------------------------+
        | ssp370     | ``----------------222222222222`` |
        +------------+----------------------------------+

    to:

        +----------+----------------------------------+
        | scenario | time                             |
        +==========+==================================+
        | ssp245   | ``hhhhhhhhhhhhhhhh111111111111`` |
        +----------+----------------------------------+
        | ssp370   | ``hhhhhhhhhhhhhhhh222222222222`` |
        +----------+----------------------------------+
    """
    if len(hist) > 1:
        raise ValueError("Too many values in hist scenario.")

    # Scenario dimension, and name of the historical scenario
    ((dim, _),) = hist.items()  # pylint: disable=unbalanced-dict-unpacking

    # Select historical scenario and drop it from the data
    h = da.sel(drop=True, **hist).dropna("time", how="all")
    ens = da.drop_sel(**hist)

    index = ens[dim]
    bare = ens.drop_vars(dim).dropna("time", how="all")

    return xr.concat([h, bare], dim="time").assign_coords({dim: index})


def _model_in_all_scens(
    da: xr.DataArray, dimensions: dict | None = None
) -> xr.DataArray:
    """Return data with only simulations that have at least one member in each scenario.

    Parameters
    ----------
    da : xr.DataArray
        Input data with dimensions for time, member, model and scenario.
    dimensions : dict, optional
        Mapping from original dimension names to standard dimension names: scenario, model, member.

    Returns
    -------
    xr.DataArray
      Data for models that have values for all scenarios.

    Notes
    -----
    In the following example, model `C` would be filtered out from the data because it has no member for `ssp370`.

    +-------+--------+--------+
    | model | members         |
    +-------+-----------------+
    |       | ssp245 | ssp370 |
    +=======+========+========+
    | A     | 1,2,3  | 1,2,3  |
    +-------+--------+--------+
    | B     | 1      | 2,3    |
    +-------+--------+--------+
    | C     | 1,2,3  |        |
    +-------+--------+--------+
    """
    if dimensions is None:
        dimensions = {}

    da = da.rename(reverse_dict(dimensions))

    ok = da.notnull().any("time").any("member").all("scenario")

    return da.sel(model=ok).rename(dimensions)


def _single_member(da: xr.DataArray, dimensions: dict | None = None) -> xr.DataArray:
    """Return data for a single member per model.

    Parameters
    ----------
    da : xr.DataArray
        Input data with dimensions for time, member, model and scenario.
    dimensions : dict
        Mapping from original dimension names to standard dimension names: scenario, model, member.

    Returns
    -------
    xr.DataArray
        Data with only one member per model.

    Notes
    -----
    In the following example, the original members would be filtered to return only the first member found for each
    scenario.

    +-------+--------+--------+----+--------+--------+
    | model | member          |    | Selected        |
    +-------+-----------------+----+-----------------+
    |       | ssp245 | ssp370 |    | ssp245 | ssp370 |
    +=======+========+========+====+========+========+
    | A     | 1,2,3  | 1,2,3  |    | 1      | 1      |
    +-------+--------+--------+----+--------+--------+
    | B     | 1,2    | 2,3    |    | 1      | 2      |
    +-------+--------+--------+----+--------+--------+
    """
    if dimensions is None:
        dimensions = {}

    da = da.rename(reverse_dict(dimensions))

    # Stack by simulation specifications - drop simulations with missing values
    full = da.stack(i=("scenario", "model", "member")).dropna("i", how="any")

    # Pick first run with data
    s = full.i.to_series()
    s[:] = np.arange(len(s))
    i = s.unstack().T.min().to_list()

    out = full.isel(i=i).unstack().squeeze()
    return out.rename(dimensions)


def reverse_dict(d: dict) -> dict:
    """Reverse dictionary."""
    return {v: k for (k, v) in d.items()}



"""
SDBA Testing Utilities Module
=============================
"""

from __future__ import annotations

import collections

import numpy as np
import pandas as pd
import xarray as xr
from scipy.stats import gamma

from xclim.sdba.utils import equally_spaced_nodes

__all__ = ["cannon_2015_dist", "cannon_2015_rvs", "nancov", "series"]


def series(values, name, start="2000-01-01"):
    """Create a DataArray with time, lon and lat dimensions."""
    coords = collections.OrderedDict()
    for dim, n in zip(("time", "lon", "lat"), values.shape, strict=False):
        if dim == "time":
            coords[dim] = pd.date_range(start, periods=n, freq="D")
        else:
            coords[dim] = xr.IndexVariable(dim, np.arange(n))

    if name == "tas":
        attrs = {
            "standard_name": "air_temperature",
            "cell_methods": "time: mean within days",
            "units": "K",
            "kind": "+",
        }
    elif name == "pr":
        attrs = {
            "standard_name": "precipitation_flux",
            "cell_methods": "time: sum over day",
            "units": "kg m-2 s-1",
            "kind": "*",
        }
    else:
        raise ValueError(f"Name `{name}` not supported.")

    return xr.DataArray(
        values,
        coords=coords,
        dims=list(coords.keys()),
        name=name,
        attrs=attrs,
    )


def cannon_2015_dist():  # noqa: D103
    # ref ~ gamma(k=4, theta=7.5)  mu: 30, sigma: 15
    ref = gamma(4, scale=7.5)

    # hist ~ gamma(k=8.15, theta=3.68) mu: 30, sigma: 10.5
    hist = gamma(8.15, scale=3.68)

    # sim ~ gamma(k=16, theta=2.63) mu: 42, sigma: 10.5
    sim = gamma(16, scale=2.63)

    return ref, hist, sim


def cannon_2015_rvs(n, random=True):  # noqa: D103
    # Frozen distributions
    fd = cannon_2015_dist()

    if random:
        r = [d.rvs(n) for d in fd]
    else:
        u = equally_spaced_nodes(n, None)
        r = [d.ppf(u) for d in fd]

    return map(lambda x: series(x, "pr"), r)


def nancov(X):
    """Numpy's cov but dropping observations with NaNs."""
    X_na = np.isnan(X).any(axis=0)
    return np.cov(X[:, ~X_na])



# pylint: disable=no-value-for-parameter
"""
Adjustment Algorithms
=====================

This file defines the different steps, to be wrapped into the Adjustment objects.
"""

from __future__ import annotations

from collections.abc import Callable, Sequence

import numpy as np
import xarray as xr

from xclim import set_options
from xclim.core.units import convert_units_to, infer_context, units
from xclim.indices.stats import _fitfunc_1d  # noqa

from . import nbutils as nbu
from . import utils as u
from ._processing import _adapt_freq
from .base import Grouper, map_blocks, map_groups
from .detrending import PolyDetrend
from .processing import escore, jitter_under_thresh, reordering, standardize


def _adapt_freq_hist(ds: xr.Dataset, adapt_freq_thresh: str):
    """Adapt frequency of null values of `hist`    in order to match `ref`."""
    with units.context(infer_context(ds.ref.attrs.get("standard_name"))):
        thresh = convert_units_to(adapt_freq_thresh, ds.ref)
    dim = ["time"] + ["window"] * ("window" in ds.hist.dims)
    return _adapt_freq.func(
        xr.Dataset({"sim": ds.hist, "ref": ds.ref}), thresh=thresh, dim=dim
    ).sim_ad


@map_groups(
    af=[Grouper.PROP, "quantiles"],
    hist_q=[Grouper.PROP, "quantiles"],
    scaling=[Grouper.PROP],
)
def dqm_train(
    ds: xr.Dataset,
    *,
    dim: str,
    kind: str,
    quantiles: np.ndarray,
    adapt_freq_thresh: str | None = None,
    jitter_under_thresh_value: str | None = None,
) -> xr.Dataset:
    """Train step on one group.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset variables:
            ref : training target
            hist : training data
    dim : str
        The dimension along which to compute the quantiles.
    kind : str
        The kind of correction to compute. See :py:func:`xclim.sdba.utils.get_correction`.
    quantiles : array-like
        The quantiles to compute.
    adapt_freq_thresh : str, optional
        Threshold for frequency adaptation. See :py:class:`xclim.sdba.processing.adapt_freq` for details.
        Default is None, meaning that frequency adaptation is not performed.
    jitter_under_thresh_value : str, optional
        Threshold under which to add uniform random noise to values, a quantity with units.
        Default is None, meaning that jitter under thresh is not performed.

    Returns
    -------
    xr.Dataset
        The dataset containing the adjustment factors, the quantiles over the training data, and the scaling factor.
    """
    ds["hist"] = (
        jitter_under_thresh(ds.hist, jitter_under_thresh_value)
        if jitter_under_thresh_value
        else ds.hist
    )
    ds["hist"] = (
        _adapt_freq_hist(ds, adapt_freq_thresh) if adapt_freq_thresh else ds.hist
    )

    refn = u.apply_correction(ds.ref, u.invert(ds.ref.mean(dim), kind), kind)
    histn = u.apply_correction(ds.hist, u.invert(ds.hist.mean(dim), kind), kind)

    ref_q = nbu.quantile(refn, quantiles, dim)
    hist_q = nbu.quantile(histn, quantiles, dim)

    af = u.get_correction(hist_q, ref_q, kind)
    mu_ref = ds.ref.mean(dim)
    mu_hist = ds.hist.mean(dim)
    scaling = u.get_correction(mu_hist, mu_ref, kind=kind)

    return xr.Dataset(data_vars={"af": af, "hist_q": hist_q, "scaling": scaling})


@map_groups(
    af=[Grouper.PROP, "quantiles"],
    hist_q=[Grouper.PROP, "quantiles"],
)
def eqm_train(
    ds: xr.Dataset,
    *,
    dim: str,
    kind: str,
    quantiles: np.ndarray,
    adapt_freq_thresh: str | None = None,
    jitter_under_thresh_value: str | None = None,
) -> xr.Dataset:
    """EQM: Train step on one group.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset variables:
            ref : training target
            hist : training data
    dim : str
        The dimension along which to compute the quantiles.
    kind : str
        The kind of correction to compute. See :py:func:`xclim.sdba.utils.get_correction`.
    quantiles : array-like
        The quantiles to compute.
    adapt_freq_thresh : str, optional
        Threshold for frequency adaptation. See :py:class:`xclim.sdba.processing.adapt_freq` for details.
        Default is None, meaning that frequency adaptation is not performed.
    jitter_under_thresh_value : str, optional
        Threshold under which to add uniform random noise to values, a quantity with units.
        Default is None, meaning that jitter under thresh is not performed.

    Returns
    -------
    xr.Dataset
        The dataset containing the adjustment factors and the quantiles over the training data.
    """
    ds["hist"] = (
        jitter_under_thresh(ds.hist, jitter_under_thresh_value)
        if jitter_under_thresh_value
        else ds.hist
    )
    ds["hist"] = (
        _adapt_freq_hist(ds, adapt_freq_thresh) if adapt_freq_thresh else ds.hist
    )
    ref_q = nbu.quantile(ds.ref, quantiles, dim)
    hist_q = nbu.quantile(ds.hist, quantiles, dim)

    af = u.get_correction(hist_q, ref_q, kind)

    return xr.Dataset(data_vars={"af": af, "hist_q": hist_q})


def _npdft_train(ref, hist, rots, quantiles, method, extrap, n_escore, standardize):
    r"""Npdf transform to correct a source `hist` into target `ref`.

    Perform a rotation, bias correct `hist` into `ref` with QuantileDeltaMapping, and rotate back.
    Do this iteratively over all rotations `rots` and conserve adjustment factors `af_q` in each iteration.

    Notes
    -----
    This function expects numpy inputs. The input arrays `ref,hist` are expected to be 2-dimensional arrays with shape:
    `(len(nfeature), len(time))`, where `nfeature` is the dimension which is mixed by the multivariate bias adjustment
    (e.g. a `multivar` dimension), i.e. `pts_dims[0]` in :py:func:`mbcn_train`. `rots` are rotation matrices with shape
    `(len(iterations), len(nfeature), len(nfeature))`.
    """
    if standardize:
        ref = (ref - np.nanmean(ref, axis=-1, keepdims=True)) / (
            np.nanstd(ref, axis=-1, keepdims=True)
        )
        hist = (hist - np.nanmean(hist, axis=-1, keepdims=True)) / (
            np.nanstd(hist, axis=-1, keepdims=True)
        )
    af_q = np.zeros((len(rots), ref.shape[0], len(quantiles)))
    escores = np.zeros(len(rots)) * np.nan
    if n_escore > 0:
        ref_step, hist_step = (
            int(np.ceil(arr.shape[1] / n_escore)) for arr in [ref, hist]
        )
    for ii, _rot in enumerate(rots):
        rot = _rot if ii == 0 else _rot @ rots[ii - 1].T
        ref, hist = rot @ ref, rot @ hist
        # loop over variables
        for iv in range(ref.shape[0]):
            ref_q, hist_q = nbu._quantile(ref[iv], quantiles), nbu._quantile(
                hist[iv], quantiles
            )
            af_q[ii, iv] = ref_q - hist_q
            af = u._interp_on_quantiles_1D(
                u._rank_bn(hist[iv]),
                quantiles,
                af_q[ii, iv],
                method=method,
                extrap=extrap,
            )
            hist[iv] = hist[iv] + af
        if n_escore > 0:
            escores[ii] = nbu._escore(ref[:, ::ref_step], hist[:, ::hist_step])
    hist = rots[-1].T @ hist
    return af_q, escores


def mbcn_train(
    ds: xr.Dataset,
    rot_matrices: xr.DataArray,
    pts_dims: Sequence[str],
    quantiles: np.ndarray,
    gw_idxs: xr.DataArray,
    interp: str,
    extrapolation: str,
    n_escore: int,
) -> xr.Dataset:
    """Npdf transform training.

    Adjusting factors obtained for each rotation in the npdf transform and conserved to be applied in
    the adjusting step in :py:func:`mcbn_adjust`.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset variables:
            ref : training target
            hist : training data
    rot_matrices : xr.DataArray
        The rotation matrices as a 3D array ('iterations', <pts_dims[0]>, <pts_dims[1]>), with shape (n_iter, <N>, <N>).
    pts_dims : sequence of str
        The name of the "multivariate" dimension and its primed counterpart. Defaults to "multivar", which
        is the normal case when using :py:func:`xclim.sdba.base.stack_variables`, and "multivar_prime",
    quantiles : array-like
        The quantiles to compute.
    gw_idxs : xr.DataArray
        Indices of the times in each windowed time group
    interp : str
        The interpolation method to use.
    extrapolation : str
        The extrapolation method to use.
    n_escore : int
        Number of elements to include in the e_score test (0 for all, < 0 to skip)

    Returns
    -------
    xr.Dataset
        The dataset containing the adjustment factors and the quantiles over the training data
        (only the npdf transform of mbcn).
    """
    # unpack data
    ref = ds.ref
    hist = ds.hist
    gr_dim = gw_idxs.attrs["group_dim"]

    # npdf training core
    af_q_l = []
    escores_l = []

    # loop over time blocks
    for ib in range(gw_idxs[gr_dim].size):
        # indices in a given time block
        indices = gw_idxs[{gr_dim: ib}].fillna(-1).astype(int).values
        ind = indices[indices >= 0]

        # npdft training : multiple rotations on standardized datasets
        # keep track of adjustment factors in each rotation for later use
        af_q, escores = xr.apply_ufunc(
            _npdft_train,
            ref[{"time": ind}],
            hist[{"time": ind}],
            rot_matrices,
            quantiles,
            input_core_dims=[
                [pts_dims[0], "time"],
                [pts_dims[0], "time"],
                ["iterations", pts_dims[1], pts_dims[0]],
                ["quantiles"],
            ],
            output_core_dims=[
                ["iterations", pts_dims[1], "quantiles"],
                ["iterations"],
            ],
            dask="parallelized",
            output_dtypes=[hist.dtype, hist.dtype],
            kwargs={
                "method": interp,
                "extrap": extrapolation,
                "n_escore": n_escore,
                "standardize": True,
            },
            vectorize=True,
        )
        af_q_l.append(af_q.expand_dims({gr_dim: [ib]}))
        escores_l.append(escores.expand_dims({gr_dim: [ib]}))
    af_q = xr.concat(af_q_l, dim=gr_dim)
    escores = xr.concat(escores_l, dim=gr_dim)
    out = xr.Dataset({"af_q": af_q, "escores": escores}).assign_coords(
        {"quantiles": quantiles, gr_dim: gw_idxs[gr_dim].values}
    )
    return out


def _npdft_adjust(sim, af_q, rots, quantiles, method, extrap):
    """Npdf transform adjusting.

    Adjusting factors `af_q` obtained in the training step are applied on the simulated data `sim` at each iterated
    rotation, see :py:func:`_npdft_train`.

    This function expects numpy inputs. `sim` can be a 2-d array with shape: `(len(nfeature), len(time))`, or
    a 3-d array with shape: `(len(period), len(nfeature), len(time))`, allowing to adjust multiple climatological periods
    all at once. `nfeature` is the dimension which is mixed by the multivariate bias adjustment
    (e.g. a `multivar` dimension), i.e. `pts_dims[0]` in :py:func:`mbcn_train`. `rots` are rotation matrices with shape
    `(len(iterations), len(nfeature), len(nfeature))`.
    """
    # add dummy dim  if period_dim absent to uniformize the function below
    # This could be done at higher level, not sure where is best
    if dummy_dim_added := (len(sim.shape) == 2):
        sim = sim[:, np.newaxis, :]

    # adjust npdft
    for ii, _rot in enumerate(rots):
        rot = _rot if ii == 0 else _rot @ rots[ii - 1].T
        sim = np.einsum("ij,j...->i...", rot, sim)
        # loop over variables
        for iv in range(sim.shape[0]):
            af = u._interp_on_quantiles_1D_multi(
                u._rank_bn(sim[iv], axis=-1),
                quantiles,
                af_q[ii, iv],
                method=method,
                extrap=extrap,
            )
            sim[iv] = sim[iv] + af

    rot = rots[-1].T
    sim = np.einsum("ij,j...->i...", rot, sim)
    if dummy_dim_added:
        sim = sim[:, 0, :]

    return sim


def mbcn_adjust(
    ref: xr.Dataset,
    hist: xr.Dataset,
    sim: xr.Dataset,
    ds: xr.Dataset,
    pts_dims: Sequence[str],
    interp: str,
    extrapolation: str,
    base: Callable,
    base_kws_vars: dict,
    adj_kws: dict,
    period_dim: str | None,
) -> xr.DataArray:
    """Perform the adjustment portion MBCn multivariate bias correction technique.

    The function :py:func:`mbcn_train` pre-computes the adjustment factors for each rotation
    in the npdf portion of the MBCn algorithm. The rest of adjustment is performed here
    in `mbcn_adjust``.

    Parameters
    ----------
    ref : xr.DataArray
        training target
    hist : xr.DataArray
        training data
    sim : xr.DataArray
        data to adjust (stacked with multivariate dimension)
    ds : xr.Dataset
        Dataset variables:
            rot_matrices : Rotation matrices used in the training step.
            af_q : Adjustment factors obtained in the training step for the npdf transform
            g_idxs : Indices of the times in each time group
            gw_idxs: Indices of the times in each windowed time group
    pts_dims : [str, str]
        The name of the "multivariate" dimension and its primed counterpart. Defaults to "multivar", which
        is the normal case when using :py:func:`xclim.sdba.base.stack_variables`, and "multivar_prime"
    interp : str
        Interpolation method for the npdf transform (same as in the training step)
    extrapolation : str
        Extrapolation method for the npdf transform (same as in the training step)
    base : BaseAdjustment
        Bias-adjustment class used for the univariate bias correction.
    base_kws_vars : Dict
        Options for univariate training for the scenario that is reordered with the output of npdf transform.
        The arguments are those expected by TrainAdjust classes along with
        - kinds : Dict of correction kinds for each variable (e.g. {"pr":"*", "tasmax":"+"})
    adj_kws : Dict
        Options for univariate adjust for the scenario that is reordered with the output of npdf transform
    period_dim : str, optional
        Name of the period dimension used when stacking time periods of `sim`  using :py:func:`xclim.core.calendar.stack_periods`.
        If specified, the interpolation of the npdf transform is performed only once and applied on all periods simultaneously.
        This should be more performant, but also more memory intensive. Defaults to `None`: No optimization will be attempted.

    Returns
    -------
    xr.Dataset
        The adjusted data.
    """
    # unpacking training parameters
    rot_matrices = ds.rot_matrices
    af_q = ds.af_q
    quantiles = af_q.quantiles
    g_idxs = ds.g_idxs
    gw_idxs = ds.gw_idxs
    gr_dim = gw_idxs.attrs["group_dim"]
    win = gw_idxs.attrs["group"][1]

    # this way of handling was letting open the possibility to perform
    # interpolation for multiple periods in the simulation all at once
    # in principle, avoiding redundancy. Need to test this on small data
    # to confirm it works,  and on big data to check performance.
    dims = ["time"] if period_dim is None else [period_dim, "time"]

    # mbcn core
    scen_mbcn = xr.zeros_like(sim)
    for ib in range(gw_idxs[gr_dim].size):
        # indices in a given time block (with and without the window)
        indices_gw = gw_idxs[{gr_dim: ib}].fillna(-1).astype(int).values
        ind_gw = indices_gw[indices_gw >= 0]
        indices_g = g_idxs[{gr_dim: ib}].fillna(-1).astype(int).values
        ind_g = indices_g[indices_g >= 0]

        # 1. univariate adjustment of sim -> scen
        # the kind may differ depending on the variables
        scen_block = xr.zeros_like(sim[{"time": ind_gw}])
        for iv, v in enumerate(sim[pts_dims[0]].values):
            sl = {"time": ind_gw, pts_dims[0]: iv}
            with set_options(sdba_extra_output=False):
                ADJ = base.train(
                    ref[sl], hist[sl], **base_kws_vars[v], skip_input_checks=True
                )
                scen_block[{pts_dims[0]: iv}] = ADJ.adjust(
                    sim[sl], **adj_kws, skip_input_checks=True
                )

        # 2. npdft adjustment of sim
        npdft_block = xr.apply_ufunc(
            _npdft_adjust,
            standardize(sim[{"time": ind_gw}].copy(), dim="time")[0],
            af_q[{gr_dim: ib}],
            rot_matrices,
            quantiles,
            input_core_dims=[
                [pts_dims[0]] + dims,
                ["iterations", pts_dims[1], "quantiles"],
                ["iterations", pts_dims[1], pts_dims[0]],
                ["quantiles"],
            ],
            output_core_dims=[
                [pts_dims[0]] + dims,
            ],
            dask="parallelized",
            output_dtypes=[sim.dtype],
            kwargs={"method": interp, "extrap": extrapolation},
            vectorize=True,
        )

        # 3. reorder scen according to npdft results
        reordered = reordering(ref=npdft_block, sim=scen_block)
        if win > 1:
            # keep  central value of window (intersecting indices in gw_idxs and g_idxs)
            scen_mbcn[{"time": ind_g}] = reordered[{"time": np.in1d(ind_gw, ind_g)}]
        else:
            scen_mbcn[{"time": ind_g}] = reordered

    return scen_mbcn.to_dataset(name="scen")


@map_blocks(reduces=[Grouper.PROP, "quantiles"], scen=[])
def qm_adjust(
    ds: xr.Dataset, *, group: Grouper, interp: str, extrapolation: str, kind: str
) -> xr.Dataset:
    """QM (DQM and EQM): Adjust step on one block.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset variables:
            af : Adjustment factors
            hist_q : Quantiles over the training data
            sim : Data to adjust.
    group : Grouper
        The grouper object.
    interp : str
        The interpolation method to use.
    extrapolation : str
        The extrapolation method to use.
    kind : str
        The kind of correction to compute. See :py:func:`xclim.sdba.utils.get_correction`.

    Returns
    -------
    xr.Dataset
        The adjusted data.
    """
    af = u.interp_on_quantiles(
        ds.sim,
        ds.hist_q,
        ds.af,
        group=group,
        method=interp,
        extrapolation=extrapolation,
    )

    scen: xr.DataArray = u.apply_correction(ds.sim, af, kind).rename("scen")
    out = scen.to_dataset()
    return out


@map_blocks(reduces=[Grouper.PROP, "quantiles"], scen=[], trend=[])
def dqm_adjust(
    ds: xr.Dataset,
    *,
    group: Grouper,
    interp: str,
    kind: str,
    extrapolation: str,
    detrend: int | PolyDetrend,
) -> xr.Dataset:
    """DQM adjustment on one block.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset variables:
            scaling : Scaling factor between ref and hist
            af : Adjustment factors
            hist_q : Quantiles over the training data
            sim : Data to adjust
    group : Grouper
        The grouper object.
    interp : str
        The interpolation method to use.
    kind : str
        The kind of correction to compute. See :py:func:`xclim.sdba.utils.get_correction`.
    extrapolation : str
        The extrapolation method to use.
    detrend : int | PolyDetrend
        The degree of the polynomial detrending to apply. If 0, no detrending is applied.

    Returns
    -------
    xr.Dataset
        The adjusted data and the trend.
    """
    scaled_sim = u.apply_correction(
        ds.sim,
        u.broadcast(
            ds.scaling,
            ds.sim,
            group=group,
            interp=interp if group.prop != "dayofyear" else "nearest",
        ),
        kind,
    )

    if isinstance(detrend, int):
        detrending = PolyDetrend(degree=detrend, kind=kind, group=group)
    else:
        detrending = detrend

    detrending = detrending.fit(scaled_sim)
    ds["sim"] = detrending.detrend(scaled_sim)
    scen = qm_adjust.func(
        ds,
        group=group,
        interp=interp,
        extrapolation=extrapolation,
        kind=kind,
    ).scen
    scen = detrending.retrend(scen)

    out = xr.Dataset({"scen": scen, "trend": detrending.ds.trend})
    return out


@map_blocks(reduces=[Grouper.PROP, "quantiles"], scen=[], sim_q=[])
def qdm_adjust(ds: xr.Dataset, *, group, interp, extrapolation, kind) -> xr.Dataset:
    """QDM: Adjust process on one block.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset variables:
            af : Adjustment factors
            hist_q : Quantiles over the training data
            sim : Data to adjust.
    """
    sim_q = group.apply(u.rank, ds.sim, main_only=True, pct=True)
    af = u.interp_on_quantiles(
        sim_q,
        ds.quantiles,
        ds.af,
        group=group,
        method=interp,
        extrapolation=extrapolation,
    )
    scen = u.apply_correction(ds.sim, af, kind)
    return xr.Dataset({"scen": scen, "sim_q": sim_q})


@map_blocks(
    reduces=[Grouper.ADD_DIMS, Grouper.DIM],
    af=[Grouper.PROP],
    hist_thresh=[Grouper.PROP],
)
def loci_train(ds: xr.Dataset, *, group, thresh) -> xr.Dataset:
    """LOCI: Train on one block.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset variables:
            ref : training target
            hist : training data
    """
    s_thresh = group.apply(
        u.map_cdf, ds.rename(hist="x", ref="y"), y_value=thresh
    ).isel(x=0)
    sth = u.broadcast(s_thresh, ds.hist, group=group)
    ws = xr.where(ds.hist >= sth, ds.hist, np.nan)
    wo = xr.where(ds.ref >= thresh, ds.ref, np.nan)

    ms = group.apply("mean", ws, skipna=True)
    mo = group.apply("mean", wo, skipna=True)

    # Adjustment factor
    af = u.get_correction(ms - s_thresh, mo - thresh, u.MULTIPLICATIVE)
    return xr.Dataset({"af": af, "hist_thresh": s_thresh})


@map_blocks(reduces=[Grouper.PROP], scen=[])
def loci_adjust(ds: xr.Dataset, *, group, thresh, interp) -> xr.Dataset:
    """LOCI: Adjust on one block.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset variables:
            hist_thresh : Hist's equivalent thresh from ref
            sim : Data to adjust
    """
    sth = u.broadcast(ds.hist_thresh, ds.sim, group=group, interp=interp)
    factor = u.broadcast(ds.af, ds.sim, group=group, interp=interp)
    with xr.set_options(keep_attrs=True):
        scen: xr.DataArray = (
            (factor * (ds.sim - sth) + thresh).clip(min=0).rename("scen")
        )
    out = scen.to_dataset()
    return out


@map_groups(af=[Grouper.PROP])
def scaling_train(ds: xr.Dataset, *, dim, kind) -> xr.Dataset:
    """Scaling: Train on one group.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset variables:
            ref : training target
            hist : training data
    """
    mhist = ds.hist.mean(dim)
    mref = ds.ref.mean(dim)
    af: xr.DataArray = u.get_correction(mhist, mref, kind).rename("af")
    out = af.to_dataset()
    return out


@map_blocks(reduces=[Grouper.PROP], scen=[])
def scaling_adjust(ds: xr.Dataset, *, group, interp, kind) -> xr.Dataset:
    """Scaling: Adjust on one block.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset variables:
            af : Adjustment factors.
            sim : Data to adjust.
    """
    af = u.broadcast(ds.af, ds.sim, group=group, interp=interp)
    scen: xr.DataArray = u.apply_correction(ds.sim, af, kind).rename("scen")
    out = scen.to_dataset()
    return out


def npdf_transform(ds: xr.Dataset, **kwargs) -> xr.Dataset:
    r"""N-pdf transform : Iterative univariate adjustment in random rotated spaces.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset variables:
            ref : Reference multivariate timeseries
            hist : simulated timeseries on the reference period
            sim : Simulated timeseries on the projected period.
            rot_matrices : Random rotation matrices.
    \*\*kwargs
        pts_dim : multivariate dimension name
        base : Adjustment class
        base_kws : Kwargs for initialising the adjustment object
        adj_kws : Kwargs of the `adjust` call
        n_escore : Number of elements to include in the e_score test (0 for all, < 0 to skip)

    Returns
    -------
    xr.Dataset
        Dataset with `scenh`, `scens` and `escores` DataArrays, where `scenh` and `scens` are `hist` and `sim`
        respectively after adjustment according to `ref`. If `n_escore` is negative, `escores` will be filled with nans.
    """
    ref = ds.ref.rename(time_hist="time")
    hist = ds.hist.rename(time_hist="time")
    sim = ds.sim
    dim = kwargs["pts_dim"]

    escores = []
    for i, R in enumerate(ds.rot_matrices.transpose("iterations", ...)):
        # @ operator stands for matrix multiplication (along named dimensions): x@R = R@x
        # @R rotates an array defined over dimension x unto new dimension x'. x@R = x'
        refp = ref @ R
        histp = hist @ R
        simp = sim @ R

        # Perform univariate adjustment in rotated space (x')
        ADJ = kwargs["base"].train(
            refp, histp, **kwargs["base_kws"], skip_input_checks=True
        )
        scenhp = ADJ.adjust(histp, **kwargs["adj_kws"], skip_input_checks=True)
        scensp = ADJ.adjust(simp, **kwargs["adj_kws"], skip_input_checks=True)

        # Rotate back to original dimension x'@R = x
        # Note that x'@R is a back rotation because the matrix multiplication is now done along x' due to xarray
        # operating along named dimensions.
        # In normal linear algebra, this is equivalent to taking @R.T, the back rotation.
        hist = scenhp @ R
        sim = scensp @ R

        # Compute score
        if kwargs["n_escore"] >= 0:
            escores.append(
                escore(
                    ref,
                    hist,
                    dims=(dim, "time"),
                    N=kwargs["n_escore"],
                    scale=True,
                ).expand_dims(iterations=[i])
            )

    if kwargs["n_escore"] >= 0:
        escores = xr.concat(escores, "iterations")
    else:
        # All nan, but with the proper shape.
        escores = (
            ref.isel({dim: 0, "time": 0}) * hist.isel({dim: 0, "time": 0})
        ).expand_dims(iterations=ds.iterations) * np.nan

    return xr.Dataset(
        data_vars={
            "scenh": hist.rename(time="time_hist").transpose(*ds.hist.dims),
            "scen": sim.transpose(*ds.sim.dims),
            "escores": escores,
        }
    )


def _fit_on_cluster(data, thresh, dist, cluster_thresh):
    """Extract clusters on 1D data and fit "dist" on the maximums."""
    _, _, _, maximums = u.get_clusters_1d(data, thresh, cluster_thresh)
    params = list(
        _fitfunc_1d(maximums - thresh, dist=dist, floc=0, nparams=3, method="ML")
    )
    # We forced 0, put back thresh.
    params[-2] = thresh
    return params


def _extremes_train_1d(ref, hist, ref_params, *, q_thresh, cluster_thresh, dist, N):
    """Train for method ExtremeValues, only for 1D input along time."""
    # Fast-track, do nothing for all-nan slices
    if all(np.isnan(ref)) or all(np.isnan(hist)):
        return np.full(N, np.nan), np.full(N, np.nan), np.nan

    # Find quantile q_thresh
    thresh = (
        np.nanquantile(ref[ref >= cluster_thresh], q_thresh)
        + np.nanquantile(hist[hist >= cluster_thresh], q_thresh)
    ) / 2

    # Fit genpareto on cluster maximums on ref (if needed) and hist.
    if np.isnan(ref_params).all():
        ref_params = _fit_on_cluster(ref, thresh, dist, cluster_thresh)

    hist_params = _fit_on_cluster(hist, thresh, dist, cluster_thresh)

    # Find probabilities of extremes according to fitted dist
    Px_ref = dist.cdf(ref[ref >= thresh], *ref_params)
    hist = hist[hist >= thresh]
    Px_hist = dist.cdf(hist, *hist_params)

    # Find common probabilities range.
    Pmax = min(Px_ref.max(), Px_hist.max())
    Pmin = max(Px_ref.min(), Px_hist.min())
    Pcommon = (Px_hist <= Pmax) & (Px_hist >= Pmin)
    Px_hist = Px_hist[Pcommon]

    # Find values of hist extremes if they followed ref's distribution.
    hist_in_ref = dist.ppf(Px_hist, *ref_params)

    # Adjustment factors, unsorted
    af = hist_in_ref / hist[Pcommon]
    # sort them in Px order, and pad to have N values.
    order = np.argsort(Px_hist)
    px_hist = np.pad(Px_hist[order], ((0, N - af.size),), constant_values=np.nan)
    af = np.pad(af[order], ((0, N - af.size),), constant_values=np.nan)

    return px_hist, af, thresh


@map_blocks(
    reduces=["time"], px_hist=["quantiles"], af=["quantiles"], thresh=[Grouper.PROP]
)
def extremes_train(
    ds: xr.Dataset,
    *,
    group: Grouper,
    q_thresh: float,
    cluster_thresh: float,
    dist,
    quantiles: np.ndarray,
) -> xr.Dataset:
    """Train extremes for a given variable series.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing the reference and historical data.
    group : Grouper
        The grouper object.
    q_thresh : float
        The quantile threshold to use.
    cluster_thresh : float
        The threshold for clustering.
    dist : Any
        The distribution to fit.
    quantiles : array-like
        The quantiles to compute.

    Returns
    -------
    xr.Dataset
        The dataset containing the quantiles, the adjustment factors, and the threshold.
    """
    px_hist, af, thresh = xr.apply_ufunc(
        _extremes_train_1d,
        ds.ref,
        ds.hist,
        ds.ref_params or np.nan,
        input_core_dims=[("time",), ("time",), ()],
        output_core_dims=[("quantiles",), ("quantiles",), ()],
        vectorize=True,
        kwargs={
            "q_thresh": q_thresh,
            "cluster_thresh": cluster_thresh,
            "dist": dist,
            "N": len(quantiles),
        },
    )
    # Outputs of map_blocks must have dimensions.
    if not isinstance(thresh, xr.DataArray):
        thresh = xr.DataArray(thresh)
    thresh = thresh.expand_dims(group=[1])
    return xr.Dataset(
        {"px_hist": px_hist, "af": af, "thresh": thresh},
        coords={"quantiles": quantiles},
    )


def _fit_cluster_and_cdf(data, thresh, dist, cluster_thresh):
    """Fit 1D cluster maximums and immediately compute CDF."""
    fut_params = _fit_on_cluster(data, thresh, dist, cluster_thresh)
    return dist.cdf(data, *fut_params)


@map_blocks(reduces=["quantiles", Grouper.PROP], scen=[])
def extremes_adjust(
    ds: xr.Dataset,
    *,
    group: Grouper,
    frac: float,
    power: float,
    dist,
    interp: str,
    extrapolation: str,
    cluster_thresh: float,
) -> xr.Dataset:
    """Adjust extremes to reflect many distribution factors.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing the reference and historical data.
    group : Grouper
        The grouper object.
    frac : float
        The fraction of the transition function.
    power : float
        The power of the transition function.
    dist : Any
        The distribution to fit.
    interp : str
        The interpolation method to use.
    extrapolation : str
        The extrapolation method to use.
    cluster_thresh : float
        The threshold for clustering.

    Returns
    -------
    xr.Dataset
        The dataset containing the adjusted data.
    """
    # Find probabilities of extremes of fut according to its own cluster-fitted dist.
    px_fut = xr.apply_ufunc(
        _fit_cluster_and_cdf,
        ds.sim,
        ds.thresh,
        input_core_dims=[["time"], []],
        output_core_dims=[["time"]],
        kwargs={"dist": dist, "cluster_thresh": cluster_thresh},
        vectorize=True,
    )

    # Find factors by interpolating from hist probs to fut probs. apply them.
    af = u.interp_on_quantiles(
        px_fut, ds.px_hist, ds.af, method=interp, extrapolation=extrapolation
    )
    scen = u.apply_correction(ds.sim, af, "*")

    # Smooth transition function between simulation and scenario.
    transition = (
        ((ds.sim - ds.thresh) / ((ds.sim.max("time")) - ds.thresh)) / frac
    ) ** power
    transition = transition.clip(0, 1)

    adjusted: xr.DataArray = (transition * scen) + ((1 - transition) * ds.scen)
    out = adjusted.rename("scen").squeeze("group", drop=True).to_dataset()
    return out


def _otc_adjust(
    X: np.ndarray,
    Y: np.ndarray,
    bin_width: dict | float | np.ndarray | None = None,
    bin_origin: dict | float | np.ndarray | None = None,
    num_iter_max: int | None = 100_000_000,
    jitter_inside_bins: bool = True,
    normalization: str | None = "max_distance",
):
    """Optimal Transport Correction of the bias of X with respect to Y.

    Parameters
    ----------
    X : np.ndarray
        Historical data to be corrected.
    Y : np.ndarray
        Bias correction reference, target of optimal transport.
    bin_width : dict or float or np.ndarray, optional
        Bin widths for specified dimensions.
    bin_origin : dict or float or np.ndarray, optional
        Bin origins for specified dimensions.
    num_iter_max : int, optional
        Maximum number of iterations used in the earth mover distance algorithm.
    jitter_inside_bins : bool
        If `False`, output points are located at the center of their bin.
        If `True`, a random location is picked uniformly inside their bin. Default is `True`.
    normalization : {None, 'standardize', 'max_distance', 'max_value'}
        Per-variable transformation applied before the distances are calculated
        in the optimal transport.

    Returns
    -------
    np.ndarray
        Adjusted data

    References
    ----------
    :cite:cts:`sdba-robin_2021`
    """
    # Initialize parameters
    if bin_width is None:
        bin_width = u.bin_width_estimator([Y, X])
    elif isinstance(bin_width, dict):
        _bin_width = u.bin_width_estimator([Y, X])
        for k, v in bin_width.items():
            _bin_width[k] = v
        bin_width = _bin_width
    elif isinstance(bin_width, float | int):
        bin_width = np.ones(X.shape[1]) * bin_width

    if bin_origin is None:
        bin_origin = np.zeros(X.shape[1])
    elif isinstance(bin_origin, dict):
        _bin_origin = np.zeros(X.shape[1])
        if bin_origin is not None:
            for v, k in bin_origin.items():
                _bin_origin[v] = k
        bin_origin = _bin_origin
    elif isinstance(bin_origin, float | int):
        bin_origin = np.ones(X.shape[1]) * bin_origin

    num_iter_max = 100_000_000 if num_iter_max is None else num_iter_max

    # Get the bin positions and frequencies of X and Y, and for all Xs the bin to which they belong
    gridX, muX, binX = u.histogram(X, bin_width, bin_origin)
    gridY, muY, _ = u.histogram(Y, bin_width, bin_origin)

    # Compute the optimal transportation plan
    plan = u.optimal_transport(gridX, gridY, muX, muY, num_iter_max, normalization)

    gridX = np.floor((gridX - bin_origin) / bin_width)
    gridY = np.floor((gridY - bin_origin) / bin_width)

    # regroup the indices of all the points belonging to a same bin
    binX_sort = np.lexsort(binX[:, ::-1].T)
    sorted_bins = binX[binX_sort]
    _, binX_start, binX_count = np.unique(
        sorted_bins, return_index=True, return_counts=True, axis=0
    )
    binX_start_sort = np.sort(binX_start)
    binX_groups = np.split(binX_sort, binX_start_sort[1:])

    out = np.empty(X.shape)
    rng = np.random.default_rng()
    # The plan row corresponding to a source bin indicates its probabilities to be transported to every target bin
    for i, binX_group in enumerate(binX_groups):
        # Pick as much target bins for this source bin as there are points in the source bin
        choice = rng.choice(range(muY.size), p=plan[i, :], size=binX_count[i])
        out[binX_group] = (gridY[choice] + 1 / 2) * bin_width + bin_origin

    if jitter_inside_bins:
        out += np.random.uniform(low=-bin_width / 2, high=bin_width / 2, size=out.shape)

    return out


@map_groups(scen=[Grouper.DIM])
def otc_adjust(
    ds: xr.Dataset,
    dim: list,
    pts_dim: str,
    bin_width: dict | float | None = None,
    bin_origin: dict | float | None = None,
    num_iter_max: int | None = 100_000_000,
    jitter_inside_bins: bool = True,
    adapt_freq_thresh: dict | None = None,
    normalization: str | None = "max_distance",
):
    """Optimal Transport Correction of the bias of `hist` with respect to `ref`.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset variables:
            ref : training target
            hist : training data
    dim : list
        The dimensions defining the distribution on which optimal transport is performed.
    pts_dim : str
        The dimension defining the multivariate components of the distribution.
    bin_width : dict or float, optional
        Bin widths for specified dimensions.
    bin_origin : dict or float, optional
        Bin origins for specified dimensions.
    num_iter_max : int, optional
        Maximum number of iterations used in the earth mover distance algorithm.
    jitter_inside_bins : bool
        If `False`, output points are located at the center of their bin.
        If `True`, a random location is picked uniformly inside their bin. Default is `True`.
    adapt_freq_thresh : dict, optional
        Threshold for frequency adaptation per variable.
    normalization : {None, 'standardize', 'max_distance', 'max_value'}
        Per-variable transformation applied before the distances are calculated
        in the optimal transport.

    Returns
    -------
    xr.Dataset
        Adjusted data
    """
    ref = ds.ref
    hist = ds.hist

    if adapt_freq_thresh is not None:
        for var, thresh in adapt_freq_thresh.items():
            hist.loc[var] = _adapt_freq_hist(
                xr.Dataset(
                    {"ref": ref.sel({pts_dim: var}), "hist": hist.sel({pts_dim: var})}
                ),
                thresh,
            )

    ref_map = {d: f"ref_{d}" for d in dim}
    ref = ref.rename(ref_map).stack(dim_ref=ref_map.values()).dropna(dim="dim_ref")

    hist = hist.stack(dim_hist=dim).dropna(dim="dim_hist")

    if isinstance(bin_width, dict):
        bin_width = {
            np.where(ref[pts_dim].values == var)[0][0]: op
            for var, op in bin_width.items()
        }
    if isinstance(bin_origin, dict):
        bin_origin = {
            np.where(ref[pts_dim].values == var)[0][0]: op
            for var, op in bin_origin.items()
        }

    scen = xr.apply_ufunc(
        _otc_adjust,
        hist,
        ref,
        kwargs={
            "bin_width": bin_width,
            "bin_origin": bin_origin,
            "num_iter_max": num_iter_max,
            "jitter_inside_bins": jitter_inside_bins,
            "normalization": normalization,
        },
        input_core_dims=[["dim_hist", pts_dim], ["dim_ref", pts_dim]],
        output_core_dims=[["dim_hist", pts_dim]],
        keep_attrs=True,
        vectorize=True,
    )

    # Pad dim differences with NA to please map_blocks
    ref = ref.unstack().rename({v: k for k, v in ref_map.items()})
    scen = scen.unstack().rename("scen")
    for d in dim:
        full_d = xr.concat([ref[d], scen[d]], dim=d).drop_duplicates(d)
        scen = scen.reindex({d: full_d})

    return scen.to_dataset()


def _dotc_adjust(
    X1: np.ndarray,
    Y0: np.ndarray,
    X0: np.ndarray,
    bin_width: dict | float | None = None,
    bin_origin: dict | float | None = None,
    num_iter_max: int | None = 100_000_000,
    cov_factor: str | None = "std",
    jitter_inside_bins: bool = True,
    kind: dict | None = None,
    normalization: str | None = "max_distance",
):
    """Dynamical Optimal Transport Correction of the bias of X with respect to Y.

    Parameters
    ----------
    X1 : np.ndarray
        Simulation data to adjust.
    Y0 : np.ndarray
        Bias correction reference.
    X0 : np.ndarray
        Historical simulation data.
    bin_width : dict or float, optional
        Bin widths for specified dimensions.
    bin_origin : dict or float, optional
        Bin origins for specified dimensions.
    num_iter_max : int, optional
        Maximum number of iterations used in the earth mover distance algorithm.
    cov_factor : str, optional
        Rescaling factor.
    jitter_inside_bins : bool
        If `False`, output points are located at the center of their bin.
        If `True`, a random location is picked uniformly inside their bin. Default is `True`.
    kind : dict, optional
        Keys are variable names and values are adjustment kinds, either additive or multiplicative.
        Unspecified dimensions are treated as "+".
    normalization : {None, 'standardize', 'max_distance', 'max_value'}
        Per-variable transformation applied before the distances are calculated
        in the optimal transport.

    Returns
    -------
    np.ndarray
        Adjusted data

    References
    ----------
    :cite:cts:`sdba-robin_2021`
    """
    # Initialize parameters
    if isinstance(bin_width, dict):
        _bin_width = u.bin_width_estimator([Y0, X0, X1])
        for v, k in bin_width.items():
            _bin_width[v] = k
        bin_width = _bin_width
    elif isinstance(bin_width, float | int):
        bin_width = np.ones(X0.shape[1]) * bin_width

    if isinstance(bin_origin, dict):
        _bin_origin = np.zeros(X0.shape[1])
        for v, k in bin_origin.items():
            _bin_origin[v] = k
        bin_origin = _bin_origin
    elif isinstance(bin_origin, float | int):
        bin_origin = np.ones(X0.shape[1]) * bin_origin

    # Map ref to hist
    yX0 = _otc_adjust(
        Y0,
        X0,
        bin_width=bin_width,
        bin_origin=bin_origin,
        num_iter_max=num_iter_max,
        jitter_inside_bins=False,
        normalization=normalization,
    )

    # Map hist to sim
    yX1 = _otc_adjust(
        yX0,
        X1,
        bin_width=bin_width,
        bin_origin=bin_origin,
        num_iter_max=num_iter_max,
        jitter_inside_bins=False,
        normalization=normalization,
    )

    # Temporal evolution
    motion = np.empty(yX0.shape)
    for j in range(yX0.shape[1]):
        if kind is not None and j in kind.keys() and kind[j] == "*":
            motion[:, j] = yX1[:, j] / yX0[:, j]
        else:
            motion[:, j] = yX1[:, j] - yX0[:, j]

    # Apply a variance dependent rescaling factor
    if cov_factor == "cholesky":
        fact0 = u.eps_cholesky(np.cov(Y0, rowvar=False))
        fact1 = u.eps_cholesky(np.cov(X0, rowvar=False))
        motion = (fact0 @ np.linalg.inv(fact1) @ motion.T).T
    elif cov_factor == "std":
        fact0 = np.std(Y0, axis=0)
        fact1 = np.std(X0, axis=0)
        motion = motion @ np.diag(fact0 / fact1)

    # Apply the evolution to ref
    Y1 = np.empty(yX0.shape)
    for j in range(yX0.shape[1]):
        if kind is not None and j in kind.keys() and kind[j] == "*":
            Y1[:, j] = Y0[:, j] * motion[:, j]
        else:
            Y1[:, j] = Y0[:, j] + motion[:, j]

    # Map sim to the evolution of ref
    Z1 = _otc_adjust(
        X1,
        Y1,
        bin_width=bin_width,
        bin_origin=bin_origin,
        num_iter_max=num_iter_max,
        jitter_inside_bins=jitter_inside_bins,
        normalization=normalization,
    )

    return Z1


@map_groups(scen=[Grouper.DIM])
def dotc_adjust(
    ds: xr.Dataset,
    dim: list,
    pts_dim: str,
    bin_width: dict | float | None = None,
    bin_origin: dict | float | None = None,
    num_iter_max: int | None = 100_000_000,
    cov_factor: str | None = "std",
    jitter_inside_bins: bool = True,
    kind: dict | None = None,
    adapt_freq_thresh: dict | None = None,
    normalization: str | None = "max_distance",
):
    """Dynamical Optimal Transport Correction of the bias of X with respect to Y.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset variables:
            ref : training target
            hist : training data
            sim : simulated data
    dim : list
        The dimensions defining the distribution on which optimal transport is performed.
    pts_dim : str
        The dimension defining the multivariate components of the distribution.
    bin_width : dict or float, optional
        Bin widths for specified dimensions.
    bin_origin : dict or float, optional
        Bin origins for specified dimensions.
    num_iter_max : int, optional
        Maximum number of iterations used in the earth mover distance algorithm.
    cov_factor : str, optional
        Rescaling factor.
    jitter_inside_bins : bool
        If `False`, output points are located at the center of their bin.
        If `True`, a random location is picked uniformly inside their bin. Default is `True`.
    kind : dict, optional
        Keys are variable names and values are adjustment kinds, either additive or multiplicative.
        Unspecified dimensions are treated as "+".
    adapt_freq_thresh : dict, optional
        Threshold for frequency adaptation per variable.
    normalization : {None, 'standardize', 'max_distance', 'max_value'}
        Per-variable transformation applied before the distances are calculated
        in the optimal transport.

    Returns
    -------
    xr.Dataset
        Adjusted data
    """
    hist = ds.hist
    sim = ds.sim
    ref = ds.ref

    if adapt_freq_thresh is not None:
        for var, thresh in adapt_freq_thresh.items():
            hist.loc[var] = _adapt_freq_hist(
                xr.Dataset(
                    {"ref": ref.sel({pts_dim: var}), "hist": hist.sel({pts_dim: var})}
                ),
                thresh,
            )

    # Drop data added by map_blocks and prepare for apply_ufunc
    hist_map = {d: f"hist_{d}" for d in dim}
    hist = (
        hist.rename(hist_map).stack(dim_hist=hist_map.values()).dropna(dim="dim_hist")
    )

    ref_map = {d: f"ref_{d}" for d in dim}
    ref = ref.rename(ref_map).stack(dim_ref=ref_map.values()).dropna(dim="dim_ref")

    sim = sim.stack(dim_sim=dim).dropna(dim="dim_sim")

    if kind is not None:
        kind = {
            np.where(ref[pts_dim].values == var)[0][0]: op for var, op in kind.items()
        }
    if isinstance(bin_width, dict):
        bin_width = {
            np.where(ref[pts_dim].values == var)[0][0]: op
            for var, op in bin_width.items()
        }
    if isinstance(bin_origin, dict):
        bin_origin = {
            np.where(ref[pts_dim].values == var)[0][0]: op
            for var, op in bin_origin.items()
        }

    scen = xr.apply_ufunc(
        _dotc_adjust,
        sim,
        ref,
        hist,
        kwargs={
            "bin_width": bin_width,
            "bin_origin": bin_origin,
            "num_iter_max": num_iter_max,
            "cov_factor": cov_factor,
            "jitter_inside_bins": jitter_inside_bins,
            "kind": kind,
            "normalization": normalization,
        },
        input_core_dims=[
            ["dim_sim", pts_dim],
            ["dim_ref", pts_dim],
            ["dim_hist", pts_dim],
        ],
        output_core_dims=[["dim_sim", pts_dim]],
        keep_attrs=True,
        vectorize=True,
    )

    # Pad dim differences with NA to please map_blocks
    hist = hist.unstack().rename({v: k for k, v in hist_map.items()})
    ref = ref.unstack().rename({v: k for k, v in ref_map.items()})
    scen = scen.unstack().rename("scen")
    for d in dim:
        full_d = xr.concat([hist[d], ref[d], scen[d]], dim=d).drop_duplicates(d)
        scen = scen.reindex({d: full_d})

    return scen.to_dataset()



from __future__ import annotations

import os

import cftime
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from numpy.testing import assert_array_equal
from xarray.coding.cftimeindex import CFTimeIndex

from xclim.core.calendar import (
    adjust_doy_calendar,
    climatological_mean_doy,
    common_calendar,
    compare_offsets,
    construct_offset,
    convert_doy,
    days_since_to_doy,
    doy_to_days_since,
    ensure_cftime_array,
    get_calendar,
    max_doy,
    parse_offset,
    percentile_doy,
    stack_periods,
    time_bnds,
    unstack_periods,
)


@pytest.fixture(
    params=[dict(start="2004-01-01T12:07:01", periods=27, freq="3MS")], ids=["3MS"]
)
def time_range_kwargs(request):
    return request.param


@pytest.fixture()
def datetime_index(time_range_kwargs):
    return pd.date_range(**time_range_kwargs)


@pytest.fixture()
def cftime_index(time_range_kwargs):
    return xr.cftime_range(**time_range_kwargs)


def da(index):
    return xr.DataArray(
        np.arange(100.0, 100.0 + index.size), coords=[index], dims=["time"]
    )


@pytest.mark.parametrize("freq", ["6480h", "302431min", "23144781s"])
def test_time_bnds(freq, datetime_index, cftime_index):
    da_datetime = da(datetime_index).resample(time=freq)
    out_time = da_datetime.mean()
    da_cftime = da(cftime_index).resample(time=freq)

    cftime_bounds = time_bnds(da_cftime, freq=freq)
    cftime_starts = cftime_bounds.isel(bnds=0)
    cftime_ends = cftime_bounds.isel(bnds=1)
    cftime_starts = CFTimeIndex(cftime_starts.values).to_datetimeindex()
    cftime_ends = CFTimeIndex(cftime_ends.values).to_datetimeindex()

    # cftime resolution goes down to microsecond only, code below corrects
    # that to allow for comparison with pandas datetime
    cftime_ends += np.timedelta64(999, "ns")
    out_periods = out_time.indexes["time"].to_period(freq)
    assert_array_equal(cftime_starts, out_periods.start_time)
    assert_array_equal(cftime_ends, out_periods.end_time)


@pytest.mark.parametrize("typ", ["pd", "xr"])
def test_time_bnds_irregular(typ):
    """Test time_bnds for irregular `middle of the month` time series."""
    if typ == "xr":
        start = xr.cftime_range("1990-01-01", periods=24, freq="MS")
        # Well. xarray string parsers do not support sub-second resolution, but cftime does.
        end = xr.cftime_range(
            "1990-01-01T23:59:59", periods=24, freq="ME"
        ) + pd.Timedelta(0.999999, "s")
    elif typ == "pd":
        start = pd.date_range("1990-01-01", periods=24, freq="MS")
        end = pd.date_range("1990-01-01 23:59:59.999999999", periods=24, freq="ME")
    else:
        raise ValueError("`typ` must be 'pd' or 'xr'")

    time = start + (end - start) / 2

    bounds = time_bnds(time, freq="ME")
    bs = bounds.isel(bnds=0)
    be = bounds.isel(bnds=1)

    assert_array_equal(bs, start)
    assert_array_equal(be, end)


@pytest.mark.parametrize("use_dask", [True, False])
def test_percentile_doy(tas_series, use_dask):
    tas = tas_series(np.arange(365), start="1/1/2001")
    if use_dask:
        tas = tas.chunk(dict(time=10))
    tas = xr.concat((tas, tas), "dim0")
    p1 = percentile_doy(tas, window=5, per=50)
    assert p1.sel(dayofyear=3, dim0=0).data == 2
    assert p1.attrs["units"] == "K"


@pytest.mark.parametrize("use_dask", [True, False])
def test_percentile_doy_nan(tas_series, use_dask):
    tas = tas_series(np.arange(365), start="1/1/2001")
    if use_dask:
        tas = tas.chunk(dict(time=10))
    tas = tas.where(tas.time.dt.dayofyear != 2)
    tas = xr.concat((tas, tas), "dim0")
    pnan = percentile_doy(tas, window=5, per=50)
    assert pnan.sel(dayofyear=3, dim0=0).data == 2.5
    assert pnan.attrs["units"] == "K"


@pytest.mark.parametrize("use_dask", [True, False])
def test_percentile_doy_no_copy(tas_series, use_dask):
    tas = tas_series(np.arange(365), start="1/1/2001")
    if use_dask:
        tas = tas.chunk(dict(time=10))
    tas = xr.concat((tas, tas), "dim0")
    original_tas = tas.copy(deep=True)
    p1 = percentile_doy(tas, window=5, per=50, copy=False)
    assert p1.sel(dayofyear=3, dim0=0).data == 2
    assert p1.attrs["units"] == "K"
    assert not np.testing.assert_array_equal(original_tas, tas)


def test_percentile_doy_invalid():
    tas = xr.DataArray(
        [0, 1],
        dims=("time",),
        coords={"time": pd.date_range("2000-01-01", periods=2, freq="h")},
    )
    with pytest.raises(ValueError):
        percentile_doy(tas)


@pytest.mark.parametrize(
    "freqA,op,freqB,exp",
    [
        ("D", ">", "h", True),
        ("2YS", "<=", "QS-DEC", False),
        ("4W", "==", "3W", False),
        ("24h", "==", "D", True),
    ],
)
def test_compare_offsets(freqA, op, freqB, exp):
    assert compare_offsets(freqA, op, freqB) is exp


def test_adjust_doy_360_to_366():
    source = xr.DataArray(np.arange(360), coords=[np.arange(1, 361)], dims="dayofyear")
    time = pd.date_range("2000-01-01", "2001-12-31", freq="D")
    target = xr.DataArray(np.arange(len(time)), coords=[time], dims="time")

    out = adjust_doy_calendar(source, target)

    assert out.sel(dayofyear=1) == source.sel(dayofyear=1)
    assert out.sel(dayofyear=366) == source.sel(dayofyear=360)


def test_adjust_doy__max_93_to_max_94():
    # GIVEN
    source = xr.DataArray(np.arange(92), coords=[np.arange(152, 244)], dims="dayofyear")
    time = xr.cftime_range("2000-06-01", periods=92, freq="D", calendar="all_leap")
    target = xr.DataArray(np.arange(len(time)), coords=[time], dims="time")
    # WHEN
    out = adjust_doy_calendar(source, target)
    # THEN
    assert out[0].dayofyear == 153
    assert out[0] == source[0]
    assert out[-1].dayofyear == 244
    assert out[-1] == source[-1]


def test_adjust_doy__leap_to_noleap_djf():
    # GIVEN
    leap_source = xr.DataArray(
        np.arange(92),
        coords=[np.concatenate([np.arange(1, 61), np.arange(335, 367)])],
        dims="dayofyear",
    )
    time = xr.cftime_range("2000-12-01", periods=91, freq="D", calendar="noleap")
    no_leap_target = xr.DataArray(np.arange(len(time)), coords=[time], dims="time")
    # WHEN
    out = adjust_doy_calendar(leap_source, no_leap_target)
    # THEN
    assert out[0].dayofyear == 1
    assert out[0] == leap_source[0]
    assert 366 not in out.dayofyear
    assert out[-1].dayofyear == 365
    assert out[-1] == leap_source[-1]


def test_adjust_doy_366_to_360():
    source = xr.DataArray(np.arange(366), coords=[np.arange(1, 367)], dims="dayofyear")
    time = xr.cftime_range("2000", periods=360, freq="D", calendar="360_day")
    target = xr.DataArray(np.arange(len(time)), coords=[time], dims="time")

    out = adjust_doy_calendar(source, target)

    assert out.sel(dayofyear=1) == source.sel(dayofyear=1)
    assert out.sel(dayofyear=360) == source.sel(dayofyear=366)


@pytest.mark.parametrize(
    "file,cal,maxdoy",
    [
        (
            (
                "CanESM2_365day",
                "pr_day_CanESM2_rcp85_r1i1p1_na10kgrid_qm-moving-50bins-detrend_2095.nc",
            ),
            "noleap",
            365,
        ),
        (
            (
                "HadGEM2-CC_360day",
                "pr_day_HadGEM2-CC_rcp85_r1i1p1_na10kgrid_qm-moving-50bins-detrend_2095.nc",
            ),
            "360_day",
            360,
        ),
        (("NRCANdaily", "nrcan_canada_daily_pr_1990.nc"), "proleptic_gregorian", 366),
    ],
)
def test_get_calendar(file, cal, maxdoy, open_dataset):
    with open_dataset(os.path.join(*file)) as ds:
        out_cal = get_calendar(ds)
        assert cal == out_cal
        assert max_doy[cal] == maxdoy


@pytest.mark.parametrize(
    "obj,cal",
    [
        ([pd.Timestamp.now()], "standard"),
        (pd.Timestamp.now(), "standard"),
        (cftime.DatetimeAllLeap(2000, 1, 1), "all_leap"),
        (np.array([cftime.DatetimeNoLeap(2000, 1, 1)]), "noleap"),
        (xr.cftime_range("2000-01-01", periods=4, freq="D"), "standard"),
    ],
)
def test_get_calendar_nonxr(obj, cal):
    assert get_calendar(obj) == cal


@pytest.mark.parametrize("obj", ["astring", {"a": "dict"}, lambda x: x])
def test_get_calendar_errors(obj):
    with pytest.raises(ValueError, match="Calendar could not be inferred from object"):
        get_calendar(obj)


def test_convert_calendar_and_doy():
    doy = xr.DataArray(
        [31, 32, 336, 364.23, 365],
        dims=("time",),
        coords={
            "time": xr.date_range("2000-01-01", periods=5, freq="YS", calendar="noleap")
        },
        attrs={"is_dayofyear": 1, "calendar": "noleap"},
    )
    out = convert_doy(doy, target_cal="360_day").convert_calendar(
        "360_day", align_on="date"
    )
    # out = convert_calendar(doy, "360_day", align_on="date", doy=True)
    np.testing.assert_allclose(
        out, [30.575342, 31.561644, 331.39726, 359.240548, 360.0]
    )
    assert out.time.dt.calendar == "360_day"
    out = convert_doy(doy, target_cal="360_day", align_on="date").convert_calendar(
        "360_day", align_on="date"
    )
    np.testing.assert_array_equal(out, [np.nan, 31, 332, 360.23, np.nan])
    assert out.time.dt.calendar == "360_day"


@pytest.mark.parametrize(
    "inp,calout",
    [
        (
            xr.DataArray(
                xr.date_range("2004-01-01", "2004-01-10", freq="D"),
                dims=("time",),
                name="time",
            ),
            "standard",
        ),
        (xr.date_range("2004-01-01", "2004-01-10", freq="D"), "standard"),
        (
            xr.DataArray(xr.date_range("2004-01-01", "2004-01-10", freq="D")).values,
            "standard",
        ),
        (xr.date_range("2004-01-01", "2004-01-10", freq="D").values, "standard"),
        (
            xr.date_range("2004-01-01", "2004-01-10", freq="D", calendar="julian"),
            "julian",
        ),
    ],
)
def test_ensure_cftime_array(inp, calout):
    out = ensure_cftime_array(inp)
    assert get_calendar(out) == calout


def test_clim_mean_doy(tas_series):
    arr = tas_series(np.ones(365 * 10))
    mean, stddev = climatological_mean_doy(arr, window=1)

    assert "dayofyear" in mean.coords
    np.testing.assert_array_equal(mean.values, 1)
    np.testing.assert_array_equal(stddev.values, 0)

    arr = tas_series(np.arange(365 * 3), start="1/1/2001")
    mean, stddev = climatological_mean_doy(arr, window=3)

    np.testing.assert_array_equal(mean[1:-1], np.arange(365, 365 * 2)[1:-1])
    np.testing.assert_array_almost_equal(stddev[1:-1], 298.0223, 4)


def test_doy_to_days_since():
    # simple test
    time = xr.date_range("2020-07-01", "2022-07-01", freq="YS-JUL")
    da = xr.DataArray(
        [190, 360, 3],
        dims=("time",),
        coords={"time": time},
        attrs={"is_dayofyear": 1, "calendar": "standard"},
    )

    out = doy_to_days_since(da)
    np.testing.assert_array_equal(out, [7, 178, 186])

    assert out.attrs["units"] == "days after 07-01"
    assert "is_dayofyear" not in out.attrs

    da2 = days_since_to_doy(out)
    xr.testing.assert_identical(da, da2)

    out = doy_to_days_since(da, start="07-01")
    np.testing.assert_array_equal(out, [7, 178, 186])

    # other calendar
    out = doy_to_days_since(da, calendar="noleap")
    assert out.attrs["calendar"] == "noleap"
    np.testing.assert_array_equal(out, [8, 178, 186])

    da2 = days_since_to_doy(out)  # calendar read from attribute
    da2.attrs.pop("calendar")  # drop for identicality
    da.attrs.pop("calendar")  # drop for identicality
    xr.testing.assert_identical(da, da2)

    # with start
    time = xr.date_range("2020-12-31", "2022-12-31", freq="YE")
    da = xr.DataArray(
        [190, 360, 3],
        dims=("time",),
        coords={"time": time},
        name="da",
        attrs={"is_dayofyear": 1, "calendar": "proleptic_gregorian"},
    )

    out = doy_to_days_since(da, start="01-02")
    np.testing.assert_array_equal(out, [188, 358, 1])

    da2 = days_since_to_doy(out)  # start read from attribute
    assert da2.name == da.name
    xr.testing.assert_identical(da, da2)

    # finer freq
    time = xr.date_range("2020-01-01", "2020-03-01", freq="MS")
    da = xr.DataArray(
        [15, 33, 66],
        dims=("time",),
        coords={"time": time},
        name="da",
        attrs={"is_dayofyear": 1, "calendar": "proleptic_gregorian"},
    )

    out = doy_to_days_since(da)
    assert out.attrs["units"] == "days after time coordinate"
    np.testing.assert_array_equal(out, [14, 1, 5])

    da2 = days_since_to_doy(out)  # start read from attribute
    xr.testing.assert_identical(da, da2)


@pytest.mark.parametrize(
    "freq,em,eb,es,ea",
    [
        ("4YS-JUL", 4, "Y", True, "JUL"),
        ("ME", 1, "M", False, None),
        ("YS", 1, "Y", True, "JAN"),
        ("3YE", 3, "Y", False, "DEC"),
        ("D", 1, "D", True, None),
        ("3W", 21, "D", True, None),
    ],
)
def test_parse_offset_valid(freq, em, eb, es, ea):
    m, b, s, a = parse_offset(freq)
    assert m == em
    assert b == eb
    assert s is es
    assert a == ea


def test_parse_offset_invalid():
    # This error actually comes from pandas, but why not test it anyway
    with pytest.raises(ValueError, match="Invalid frequency"):
        parse_offset("day")


@pytest.mark.parametrize(
    "m,b,s,a,exp",
    [
        (1, "Y", True, None, "YS-JAN"),
        (2, "Q", False, "DEC", "2QE-DEC"),
        (1, "D", False, None, "D"),
    ],
)
def test_construct_offset(m, b, s, a, exp):
    assert construct_offset(m, b, s, a) == exp


@pytest.mark.parametrize(
    "inputs,join,expected",
    [
        (["noleap", "standard"], "outer", "standard"),
        (["noleap", "standard"], "inner", "noleap"),
        (["default", "default"], "outer", "default"),
        (["all_leap", "default"], "inner", "standard"),
    ],
)
def test_common_calendars(inputs, join, expected):
    assert expected == common_calendar(inputs, join=join)


def test_convert_doy():
    doy = xr.DataArray(
        [31, 32, 336, 364.23, 365],
        dims=("time",),
        coords={
            "time": xr.date_range("2000-01-01", periods=5, freq="YS", calendar="noleap")
        },
        attrs={"is_dayofyear": 1, "calendar": "noleap"},
    )

    out = convert_doy(doy, "360_day", align_on="date")
    np.testing.assert_array_equal(out, [np.nan, 31, 332, 360.23, np.nan])
    assert out.time.dt.calendar == "noleap"
    out = convert_doy(doy, "360_day", align_on="year")
    np.testing.assert_allclose(
        out, [30.575342, 31.561644, 331.39726, 359.240548, 360.0]
    )

    doy = xr.DataArray(
        [31, 200.48, 190, 60, 300.54],
        dims=("time",),
        coords={
            "time": xr.date_range(
                "2000-01-01", periods=5, freq="YS-JUL", calendar="standard"
            )
        },
        attrs={"is_dayofyear": 1, "calendar": "standard"},
    ).expand_dims(lat=[10, 20, 30])

    out = convert_doy(doy, "noleap", align_on="date")
    np.testing.assert_array_equal(out.isel(lat=0), [31, 200.48, 190, np.nan, 299.54])
    out = convert_doy(doy, "noleap", align_on="year")
    np.testing.assert_allclose(
        out.isel(lat=0), [31.0, 200.48, 190.0, 59.83607, 299.71885]
    )


@pytest.mark.parametrize("cftime", [True, False])
@pytest.mark.parametrize(
    "w,s,m,f,ss",
    [(30, 10, None, "YS", 0), (3, 1, None, "QS-DEC", 60), (6, None, None, "MS", 0)],
)
def test_stack_periods(tas_series, cftime, w, s, m, f, ss):
    da = tas_series(np.arange(365 * 50), start="2000-01-01", cftime=cftime)

    da_stck = stack_periods(
        da, window=w, stride=s, min_length=m, freq=f, align_days=False
    )

    assert "period_length" in da_stck.coords

    da2 = unstack_periods(da_stck)

    xr.testing.assert_identical(da2, da.isel(time=slice(ss, da2.time.size + ss)))


def test_stack_periods_special(tas_series):
    da = tas_series(np.arange(365 * 48 + 12), cftime=True, start="2004-01-01")

    with pytest.raises(ValueError, match="unaligned day-of-year"):
        stack_periods(da)

    da = da.convert_calendar("noleap")

    da_stck = stack_periods(da, dim="horizon")
    np.testing.assert_array_equal(da_stck.horizon_length, 10950)

    with pytest.raises(ValueError, match="can't find the window"):
        unstack_periods(da_stck)

    da2 = unstack_periods(da_stck.drop_vars("horizon_length"), dim="horizon")
    xr.testing.assert_identical(da2, da.isel(time=slice(0, da2.time.size)))



from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from xclim import atmos
from xclim.indices.fire import (
    griffiths_drought_factor,
    keetch_byram_drought_index,
    mcarthur_forest_fire_danger_index,
)

data_url = "ERA5/daily_surface_cancities_1990-1993.nc"


class TestFFDI:
    @pytest.mark.parametrize(
        "p,t,pa,k0,exp",
        [
            (10 * [100], 10 * [0], [1.0], [0.0], 0.0),
            (10 * [0], 10 * [100], [1.0], [0.0], 203.2),
            ([10, 0, 0.1, 6, 0, 0, 0.5, 0.3, 0, 1], 10 * [30], [1.0], [0.0], 7.25278),
            (10 * [0], [20, 30, 20, 30, 30, 25, 40, 35, 20, 20], [1.0], [0.0], 8.46632),
            (
                [10, 0, 0.1, 6, 0, 0, 0.5, 0.3, 0, 1],
                [20, 30, 20, 30, 30, 25, 40, 35, 20, 20],
                [1.0],
                [0.0],
                7.10174,
            ),
            (
                [10, 0, 0.1, 6, 0, 0, 0.5, 0.3, 0, 1],
                [20, 30, 20, 30, 30, 25, 40, 35, 20, 20],
                [1.0],
                [10.0],
                12.18341,
            ),
            (
                [10, 0, 0.1, 6, 0, 0, 0.5, 0.3, 0, 1],
                [20, 30, 20, 30, 30, 25, 40, 35, 20, 20],
                [100.0],
                [0.0],
                8.45569,
            ),
            (
                [10, 0, 0.1, 6, 0, 0, 0.5, 0.3, 0, 1],
                [20, 30, 20, 30, 30, 25, 40, 35, 20, 20],
                [1.0],
                [203.2],
                197.33375,
            ),
        ],
    )
    def test_keetch_byram_drought_index(
        self, p, t, pa, k0, exp, pr_series, tasmax_series
    ):
        """Compare output to calculation by hand"""
        pr = pr_series(p, units="mm/day")
        tasmax = tasmax_series(t, units="degC")
        pr_annual = xr.DataArray(pa, attrs={"units": "mm/year"})
        kbdi0 = xr.DataArray(k0, attrs={"units": "mm/day"})

        kbdi_final = keetch_byram_drought_index(pr, tasmax, pr_annual, kbdi0).isel(
            time=-1
        )
        np.testing.assert_allclose(kbdi_final, exp, atol=1e-5)

    @pytest.mark.parametrize(
        "p, s, exp, test_discrete",
        [
            (17 * [0] + [5, 10, 20], 20 * [10], 0.40471, False),
            ([20, 10, 5] + 17 * [0], 20 * [10], 6.13148, True),
            (
                [0, 30, 5, 0, 0, 5, 10, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 1, 3, 1],
                20 * [30],
                6.82454,
                True,
            ),
            (
                [0, 10, 5, 0, 0, 5, 10, 0, 0, 20, 0, 0, 0, 20, 0, 0, 0, 5, 4, 3],
                20 * [30],
                6.59186,
                False,
            ),
            (
                [0, 10, 5, 0, 0, 50, 100, 0, 0, 20, 0, 0, 0, 0, 0, 0, 0, 1, 3, 1],
                20 * [10],
                3.91578,
                False,
            ),
            (
                [0, 300, 5, 0, 0, 50, 100, 0, 0, 20, 0, 0, 0, 0, 0, 0, 0, 1, 3, 1],
                20 * [30],
                3.76635,
                False,
            ),
        ],
    )
    def test_griffiths_drought_factor(self, p, exp, s, test_discrete, pr_series):
        """Compare output for a single window to calculation by hand"""
        pr = pr_series(p, units="mm/day")
        smd = pr_series(s, units="mm/day")

        df = griffiths_drought_factor(pr, smd, "xlim").isel(time=-1)
        np.testing.assert_allclose(df, exp, atol=1e-5)

        if test_discrete:
            df = griffiths_drought_factor(pr, smd, "discrete").isel(time=-1)
            np.testing.assert_allclose(df, round(exp), atol=1e-5)

    def test_griffiths_drought_factor_sliding(self, pr_series):
        """Compare output for a simple case to calculation by hand"""
        p = np.zeros(24)
        p[19] = 20.0
        pr = pr_series(p, units="mm/day")
        smd = pr_series(20 * np.ones(24), units="mm/day")
        exp = np.array([1.07024, 3.14744, 4.71645, 5.64112, 6.14665])

        df = griffiths_drought_factor(pr, smd, "xlim").isel(time=slice(19, None))
        np.testing.assert_allclose(df, exp, atol=1e-5)

    def test_mcarthur_forest_fire_danger_index(
        self, pr_series, tasmax_series, hurs_series, sfcWind_series
    ):
        """Compare output to calculation by hand"""
        D = pr_series(range(1, 11), units="")  # This is probably not good practice?
        T = tasmax_series(range(30, 40), units="degC")
        H = hurs_series(range(10, 20))
        V = sfcWind_series(range(10, 20))

        # Compare FFDI to values calculated using original arrangement of the FFDI:
        exp = 2.0 * np.exp(
            -0.450 + 0.987 * np.log(D) - 0.0345 * H + 0.0338 * T + 0.0234 * V
        )
        ffdi = mcarthur_forest_fire_danger_index(D, T, H, V)
        np.testing.assert_allclose(ffdi, exp, rtol=1e-6)

    @pytest.mark.slow
    @pytest.mark.parametrize("init_kbdi", [True, False])
    @pytest.mark.parametrize("limiting_func", ["xlim", "discrete"])
    def test_ffdi_indicators(self, open_dataset, init_kbdi, limiting_func):
        """Test the FFDI indicators using real data"""
        # I couldn't find any high quality data or code to test against. I considered
        # the CEMS GEFF dataset, and the R packages ClimInd and ecbtools but all use
        # older definitions of the KBDI and DF that differ from our code and I don't
        # think reflect the modern literature.
        # For now, we just test that the indicators run using real data and that the
        # outputs look sensible
        test_data = open_dataset(data_url)

        pr_annual = test_data["pr"].resample(time="YS").mean().mean("time")
        pr_annual.attrs["units"] = test_data["pr"].attrs["units"]

        if init_kbdi:
            kbdi0 = xr.ones_like(pr_annual) + 203.2
            kbdi0.attrs["units"] = test_data["pr"].attrs["units"]
        else:
            kbdi0 = None

        kbdi = atmos.keetch_byram_drought_index(
            test_data["pr"], test_data["tasmax"], pr_annual, kbdi0
        )
        assert (kbdi >= 0).all()
        assert (kbdi <= 203.2).all()
        assert kbdi.shape == test_data["pr"].shape

        if limiting_func == "xlim":
            df_max = 10.7216381
        else:
            df_max = 10

        df = atmos.griffiths_drought_factor(test_data["pr"], kbdi)
        assert (df.isel(time=slice(19, None)) >= 0).all()
        assert (df.isel(time=slice(19, None)) <= df_max).all()
        assert df.shape == test_data["pr"].shape

        ffdi = atmos.mcarthur_forest_fire_danger_index(
            df, test_data["tasmax"], test_data["hurs"], test_data["sfcWindmax"]
        )
        assert (ffdi.isel(time=slice(19, None)) >= 0).all()
        assert ffdi.shape == test_data["pr"].shape



#!/usr/bin/env python
# Tests for `xclim` package, command line interface
from __future__ import annotations

import numpy as np
import pytest
import xarray as xr
from click.testing import CliRunner

import xclim
from xclim.cli import cli

try:
    from dask.distributed import Client
except ImportError:
    Client = None


K2C = 273.15


@pytest.mark.parametrize(
    "indicators,indnames",
    [
        ([xclim.atmos.tg_mean], ["tg_mean"]),
        (
            # Note: This test is dependent on indicator name length and terminal dimensions.
            [xclim.atmos.tn_mean, xclim.atmos.ice_days],
            ["tn_mean", "ice_days"],
        ),
    ],
)
def test_info(indicators, indnames):
    runner = CliRunner()
    results = runner.invoke(cli, ["info"] + indnames)

    for ind in indicators:
        assert ind.title in results.output
        assert ind.identifier in results.output


def test_indices():
    runner = CliRunner()
    results = runner.invoke(cli, ["indices"])

    for name in xclim.core.indicator.registry:
        assert name.lower() in results.output


@pytest.mark.parametrize(
    "indicator,indname",
    [
        (xclim.atmos.heating_degree_days, "heating_degree_days"),
        (xclim.land.base_flow_index, "base_flow_index"),
    ],
)
def test_indicator_help(indicator, indname):
    runner = CliRunner()
    results = runner.invoke(cli, [indname, "--help"])

    for name in indicator.parameters.keys():
        if name not in ["ds", "indexer"]:
            assert name in results.output


@pytest.mark.parametrize(
    "indicator,expected,varnames",
    [
        ("tg_mean", 272.15, ["tas"]),
        ("dtrvar", 0.0, ["tasmin", "tasmax"]),
        ("heating_degree_days", 6588.0, ["tas"]),
        ("solidprcptot", 31622400.0, ["tas", "pr"]),
    ],
)
def test_normal_computation(
    tasmin_series, tasmax_series, pr_series, tmp_path, indicator, expected, varnames
):
    tasmin = tasmin_series(np.ones(366) + 270.15, start="1/1/2000")
    tasmax = tasmax_series(np.ones(366) + 272.15, start="1/1/2000")
    pr = pr_series(np.ones(366), start="1/1/2000")
    ds = xr.Dataset(
        data_vars={
            "tasmin": tasmin,
            "tasmax": tasmax,
            "tas": xclim.atmos.tg(tasmin, tasmax),
            "pr": pr,
        }
    )
    input_file = tmp_path / "in.nc"
    output_file = tmp_path / "out.nc"

    ds.to_netcdf(input_file, engine="h5netcdf")

    args = ["-i", str(input_file), "-o", str(output_file), "-v", indicator]
    runner = CliRunner()
    results = runner.invoke(cli, args)
    for varname in varnames:
        assert f"Parsed {varname} = {varname}" in results.output
    assert "Processing :" in results.output
    assert "100% Completed" in results.output

    out = xr.open_dataset(output_file)
    outvar = list(out.data_vars.values())[0]
    np.testing.assert_allclose(outvar[0], expected)


def test_multi_input(tas_series, pr_series, tmp_path):
    tas = tas_series(np.ones(366) + 273.15, start="1/1/2000")
    pr = pr_series(np.ones(366), start="1/1/2000")
    tas_file = tmp_path / "multi_tas_in.nc"
    pr_file = tmp_path / "multi_pr_in.nc"
    output_file = tmp_path / "out.nc"

    tas.to_dataset().to_netcdf(tas_file)
    pr.to_dataset().to_netcdf(pr_file)

    runner = CliRunner()
    results = runner.invoke(
        cli,
        [
            "-i",
            str(tmp_path / "multi_*_in.nc"),
            "-o",
            str(output_file),
            "-v",
            "solidprcptot",
        ],
    )
    assert "Processing : solidprcptot" in results.output

    out = xr.open_dataset(output_file)
    assert out.solidprcptot.sum() == 0


def test_multi_output(tmp_path, open_dataset):
    ds = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc")
    input_file = tmp_path / "ws_in.nc"
    output_file = tmp_path / "out.nc"
    ds.to_netcdf(input_file, engine="h5netcdf")

    runner = CliRunner()
    results = runner.invoke(
        cli,
        [
            "-i",
            str(input_file),
            "-o",
            str(output_file),
            "-v",
            "wind_speed_from_vector",
        ],
    )
    assert "Processing : wind_speed_from_vector" in results.output


def test_renaming_variable(tas_series, tmp_path):
    tas = tas_series(np.ones(366), start="1/1/2000")
    input_file = tmp_path / "tas.nc"
    output_file = tmp_path / "out.nc"
    tas.name = "tas"
    tas.to_netcdf(input_file)
    with xclim.set_options(cf_compliance="warn"):
        runner = CliRunner()
        results = runner.invoke(
            cli,
            [
                "-i",
                str(input_file),
                "-o",
                str(output_file),
                "-v",
                "tn_mean",
                "--tasmin",
                "tas",
            ],
        )
        assert "Processing : tn_mean" in results.output
        assert "100% Completed" in results.output

    out = xr.open_dataset(output_file)
    assert out.tn_mean[0] == 1.0


def test_indicator_chain(tas_series, tmp_path):
    tas = tas_series(np.ones(366), start="1/1/2000")
    input_file = tmp_path / "tas.nc"
    output_file = tmp_path / "out.nc"

    tas.to_netcdf(input_file)

    runner = CliRunner()
    results = runner.invoke(
        cli,
        [
            "-i",
            str(input_file),
            "-o",
            str(output_file),
            "-v",
            "tg_mean",
            "growing_degree_days",
        ],
    )

    assert "Processing : tg_mean" in results.output
    assert "Processing : growing_degree_days" in results.output
    assert "100% Completed" in results.output

    out = xr.open_dataset(output_file)
    assert out.tg_mean[0] == 1.0
    assert out.growing_degree_days[0] == 0


def test_missing_variable(tas_series, tmp_path):
    tas = tas_series(np.ones(366), start="1/1/2000")
    input_file = tmp_path / "tas.nc"
    output_file = tmp_path / "out.nc"

    tas.to_netcdf(input_file, engine="h5netcdf")

    runner = CliRunner()
    results = runner.invoke(
        cli, ["-i", str(input_file), "-o", str(output_file), "tn_mean"]
    )
    assert results.exit_code == 2
    assert "'tasmin' was not found in the input dataset." in results.output


@pytest.mark.parametrize(
    "options,output",
    [
        (["--dask-nthreads", "2"], "Error: '--dask-maxmem' must be given"),
        (["--chunks", "time:90"], "100% Complete"),
        (["--chunks", "time:90,lat:5"], "100% Completed"),
        (["--version"], xclim.__version__),
    ],
)
def test_global_options(tas_series, tmp_path, options, output):
    if "dask" in options[0]:
        pytest.importorskip("dask.distributed")
    tas = tas_series(np.ones(366), start="1/1/2000")
    tas = xr.concat([tas] * 10, dim="lat")
    input_file = tmp_path / "tas.nc"
    output_file = tmp_path / "out.nc"

    tas.to_netcdf(input_file, engine="h5netcdf")

    runner = CliRunner()
    results = runner.invoke(
        cli,
        ["-i", str(input_file), "-o", str(output_file)] + options + ["tg_mean"],
    )

    assert output in results.output


def test_suspicious_precipitation_flags(pr_series, tmp_path):
    bad_pr = pr_series(np.zeros(365), start="1971-01-01")
    # Add some strangeness
    bad_pr[8] = -1e-6  # negative values
    bad_pr[120] = 301 / 3600 / 24  # 301mm/day
    bad_pr[121:141] = 1.1574074074074072e-05  # 1mm/day
    bad_pr[200:300] = 5.787037037037036e-05  # 5mm/day
    input_file = tmp_path / "bad_pr.nc"
    output_file = tmp_path / "out.nc"

    bad_pr.to_netcdf(input_file)

    runner = CliRunner()
    runner.invoke(
        cli, ["-i", str(input_file), "-o", str(output_file), "dataflags", "pr"]
    )
    with xr.open_dataset(output_file) as ds:
        for var in ds.data_vars:
            assert var


@pytest.mark.slow
def test_dataflags_output(tmp_path, tas_series, tasmax_series, tasmin_series):
    ds = xr.Dataset()
    for series, val in zip(
        [tas_series, tasmax_series, tasmin_series], [0, 10, -10], strict=False
    ):
        vals = val + K2C + np.sin(np.pi * np.arange(366 * 3) / 366)
        arr = series(vals, start="1971-01-01")
        ds = xr.merge([ds, arr])
    input_file = tmp_path / "ws_in.nc"
    ds.to_netcdf(input_file, engine="h5netcdf")

    runner = CliRunner()
    results = runner.invoke(
        cli,
        [
            "-i",
            str(input_file),
            "dataflags",
            "-r",
        ],
    )
    assert "Dataset passes quality control checks!" in results.output


def test_bad_usage(tas_series, tmp_path):
    tas = tas_series(np.ones(366), start="1/1/2000")
    input_file = tmp_path / "tas.nc"
    output_file = tmp_path / "out.nc"

    tas.to_netcdf(input_file, engine="h5netcdf")

    runner = CliRunner()

    # No command
    results = runner.invoke(cli, ["-i", str(input_file)])
    assert "Missing command" in results.output

    # Indicator not found:
    results = runner.invoke(cli, ["info", "mean_ether_velocity"])
    assert "Indicator 'mean_ether_velocity' not found in xclim" in results.output

    # No input file given
    results = runner.invoke(cli, ["-o", str(output_file), "base_flow_index"])
    assert "No input file name given" in results.output

    # No output file given
    results = runner.invoke(cli, ["-i", str(input_file), "tg_mean"])
    assert "No output file name given" in results.output

    results = runner.invoke(
        cli,
        [
            "-i",
            str(input_file),
            "-o",
            str(output_file),
            "--dask-nthreads",
            "2",
            "tg_mean",
        ],
    )
    if Client is None:  # dask.distributed not installed
        assert "distributed scheduler is not installed" in results.output
    else:
        assert "'--dask-maxmem' must be given" in results.output


@pytest.mark.requires_docs
@pytest.mark.parametrize("method, pattern", [("-r", "`GH/"), ("-m", "[GH/")])
def test_release_notes(method, pattern):
    runner = CliRunner()
    results = runner.invoke(
        cli,
        ["release_notes", method],
    )
    assert ":pull:`" not in results.output
    assert ":issue:`" not in results.output
    assert ":user:`" not in results.output
    assert pattern in results.output


@pytest.mark.parametrize(
    "method, error",
    [
        (
            ["-m", "-r"],
            "Cannot return both Markdown and ReStructuredText in same release_notes call.",
        ),
        (list(), "Must specify Markdown (-m) or ReStructuredText (-r)."),
    ],
)
def test_release_notes_failure(method, error):
    runner = CliRunner()
    results = runner.invoke(
        cli,
        ["release_notes", *method],
    )
    assert error in results.output


def test_show_version_info():
    runner = CliRunner()
    results = runner.invoke(cli, ["show_version_info"])
    assert "INSTALLED VERSIONS" in results.output
    assert "python" in results.output
    assert "boltons: installed" in results.output
